"""Numerical methods for mathematical functions neede for the program"""

#This method finds the factorial of a given number(num).
#The method calls for an integer input num
#The method returns the factorial(represented as fact in the method)
def factorial(num):
    """Finds the factorial of the input integer.
    
    Args:
        num (int): The integer to find the factorial of.

    Returns:
        fact (int): The factorial of num.
    """
    #If the number provided is zero then the factorial is 1
    if num == 0:
        fact = 1
    #Otherwise set fact to 1 and begin finding the factorial r is
    #used to find each num-n for n=0 to n=num each value in r is
    #then used to compute the factorial
    else:
        fact = 1
        r = list(range(1,num+1))
        for i in r:
            fact *= i
    return fact

def binomial_coefficient(n, k):
    """Finds the binomial coefficient n choose k. See
    https://en.wikipedia.org/wiki/Binomial_coefficient for details.

    Args:
        n (int): An integer.
        k (int): An integer.

    Returns:
       t (int): The binomial coefficient, n choose k.

    """
    if not (-1 < k < n+1):
        return 0
    if k==0 and n == 0:
        return 1
    t = 1
    if k < n-k:
        for i in range(n, n-k, -1):
            t = t*i//(n-i+1)
    else:
        for i in range(n, k, -1):
            t = t*i//(n-i+1)

    return t



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
AUTHORS: Conrad W. Rosenbrock, Wiley S. Morgan (June 2015)

Classes to support the calculation of coefficients for specific terms in a product
of multinomials. Construct a product class by specifying the exponent and target term
and then add multinomials using the Product instance's append(). The coefficient is
then available from the coeff().
"""
from functools import reduce
class Sequence(object):
    """Represents an exponent-limited sequence with a single root. Here,
    sequence represents a sequence of integer values k_1, k_2...k_j
    that are the exponents of a single term in a multinomial. The root
    of the sequence is one of the k_i; its children become sets of
    sequences including variables to the right of i.

    Attributes:
        used (int): The sum of the exponents that have already been used.
        parent (Sequence): A Sequence instance for the variable previouse to this one.
        kids (Sequence): The next possible Sequence instances.
        varcount (int): The number of variables in the polynomial.
    """
    def __init__(self, root, possibles, i, powersum, targets, parent=None):
        """Initializes a sequence collector for a variable. 'Term' refers to a product
        of variables like x^i.y^j.z^k, then x has index 0, y has 1, etc.

        Args:
            root (int): The exponent of the variable to the left in the multinomial term.
            possibles (list): A list of possible values for each variable in the multinomial.
            i (int): the index of the variable being sequenced in the term.
            powersum (int): the maximum value that the sum of exponents in the sequence 
              is allowed to have.

            parent (Sequence): A Sequence instance for the variable the *left* of this one
              (i.e. has index i-1).
        """
        self._root = root
        self.used = root + (0 if parent is None else parent.used)
        self.parent = parent

        #We only keep recursively defining sequences until we run out of variables in
        #the term. Possibles is a list of possible exponents for each variable in the
        #term and has the same number of items as variables in the term.
        if i < len(targets):
            #Filter the possible values for the variable being considered based on the
            #exponent of the multinomial. When multinomials are expanded, the sum of
            #the exponents in any term must be less than the exponent on the multinomial
            #times the maximum power of any of its (unexpanded) terms.

            #We find all the possible values for this variable by ensuring that:
            # 1) it's exponent is compatible with the exponents of all variables to the left of it.
            # 2) the exponent we are suggesting is in the list of possible values for the variable.
            # 3) the exponent remains positive.
            self.kids = [Sequence(p-root, possibles, i+1, powersum, targets, self) 
                         for p in possibles if p-root >= 0
                         and p-root <= targets[i]
                         and abs(p - root) <= powersum-self.used 
                         and abs(p-self.used) % possibles[1] == 0]
        else:
            self.kids = []

        self._kidcount = None
        self.varcount = len(targets)

    @property
    def kidcount(self): #pragma: no cover
        """Returns the number of children and grandchildren to the last generation.

        Returns:
            _kidcount (int): The number of descendents of the parent.
        """
        if self._kidcount is None:
            _kidcount = sum([k.kidcount for k in self.kids])
            if _kidcount == 0:
                _kidcount = len(self.kids)
        return _kidcount

    def expand(self, depth=0):
        """Recursively generates a list of all relevant sequences for this multinomial term.

        Args:
            depth (int, optional): The current depth in the recursion.

        Returns:
            sequences (list): The relevant sequences for the multinomial.
        """
        #Iterate through the child sequences and add their variable root values if
        #the total sequence sums to the target.
        sequences = []
        for kid in self.kids:
            for seq in kid.expand(depth+1):
                #Here is where the recursion happens; we add the sequence of this variable's
                #children to the right of this root.
                sequences.append((self._root,) + seq)

        if len(self.kids) == 0:
            if depth == self.varcount-1:
                return [(self._root,)]
            else: #pragma: no cover
                return [(self._root,) + (0,)*(self.varcount-(depth+1))]
        else:
            return sequences

    def expand_noappend(self, sequences, start, varindex): #pragma: no cover
        """Implements an expansion that doesn't use python's append.

        Args:
            sequences (list): A list of the relevant sequences for the multinomials.
            start (int): Starting index.
            varindex (int): The current variabl index.

        Returns:
            sequences (list): The relevant sequences for the multinomial.

        Raises:
            ValueError: if self.kidcount is zero.
        """
        if len(sequences) == 0:
            if self.kidcount == 0:
                raise ValueError("This can't happen!")
            sequences = [[0]*self.varcount for i in range(self.kidcount)]

        cursor = start
        for kid in self.kids:
            kid.expand_noappend(sequences, cursor, varindex+1)
            cursor += kid.kidcount

            if varindex == self.varcount-1:
                cursor += 1

            for k in range(start, cursor):
                sequences[k][varindex-1] = self._root
            start = cursor

        if self.kidcount == 0:
            sequences[cursor][varindex-1] = self._root
            
        return sequences

class Product(object):
    """Represents a product of multinomials for which only a single term is interesting.

    Attributes:
        coefficient (int): The scalar integer multiplying this product of multinomials.
        targets (list): A list of exponents for the terms in the products.
        multinoms (list): A list of multinomials.
    """
    def __init__(self, coefficient, targets):
        """Initializes the empty product of multinomials.

        Args:
            coefficient (int): The scalar integer multiplying this product of multinomials.
            targets (list): A list of exponents for the only interesting term in the 
              product. The list is in the order that the variables appear in each multinomial.
        """
        self.coefficient = coefficient
        self.targets = targets
        self.multinoms = []

    def coeff(self):
        """Returns the coefficient of the term with the target exponents if all the multinomials
        in the product were expanded and had their terms collected.

        Returns:
            coefficient (int): The coefficient of th term with the target exponents.
        """
        #If this is an isolated multinomial, we only need to check the coefficient of the target
        #term.
        if len(self.multinoms) == 1:
            if all(self.multinoms[0].power-t>0 for t in self.targets):
                return 0
            else:
                return self.multinoms[0].nchoosekm(self.targets)*self.coefficient
        
        from itertools import product
        #Get a list of the possible exponents for each variable in each of the multinomials.
        #We start with the first variable and choose only those combinations of exponents
        #across *all* the multinomials that give the correct target exponent for that variable.
        possibles = [n.possible_powers for n in self.multinoms]
        
        seq0 = [s for s in product(*possibles) if sum(s) == self.targets[0]]
        #Next, we construct Sequence instances for each of the first variable compatible
        #possibilities and follow them through to the other variables.
        coeffs = 0
        for seq in seq0:
            mnseq = []
            #Each sequence calculated from the first variable has an entry for each multinomial
            #in this product. The Sequence instances construct smart sequences for the remaining
            #variables in each multinomial separately
            len_seq = len(seq)
            for i in range(len_seq):
                varseq = Sequence(seq[i], possibles[i], 1, self.multinoms[i].powersum, self.targets)
                mnseq.append(varseq.expand())
            coeffs += self._sum_sequences(mnseq)

        return int(coeffs)*self.coefficient

    def _sum_sequences(self, mnseq):
        """Sums all the possible combinations of relevant sequences based of the variable 
        sequence lists specified.
        
        Args:
            mnseq (list): a list of possible variable sequences in each multinomial (one for 
              each multinomial) that might contribute to the correct target variable.

        Returns:
            coeffs (int): The coefficients of the sequence.
        """
        from itertools import product
        from operator import mul
        #We can also filter the sequences by enforcing the constraints that the exponents
        #correctly reproduce the target across all the multionmials in the product. Get hold
        #of all the combinations of sequences across the multinomials and check each for
        #conformance to the targets.
        coeffs = 0
        for seq in product(*mnseq):
            expsum = [sum(zs) for zs in zip(*seq)]
            if expsum == self.targets:
                coeffs += reduce(mul, [m.nchoosekm(s) for m, s in zip(self.multinoms, seq)])

        return coeffs

    def __str__(self):
        #First we need to sort the multinomials by their exponent.
        sortedmns = sorted(self.multinoms, key=(lambda m: (m.exponent,m.power)), reverse=True)
        return str(self.coefficient) + ''.join([str(mn) for mn in sortedmns])            

class Multinomial(object):
    """Represents a multinomial expansion.

    Attributes:
        power (list): The power on each of the *unexpanded* variables in the multinomial;
          of the form (x^2+y^2+z^2) => 2.

        coef (list): The coefficient on each of the variables in the multinomial; e.g.
          (x^2 + a y^2) => a. For more than two variables, each color that *has* arrowing
          gets the coefficient while the others get 1.

        arrowings (list): The exponents of the arrows terms.
        exponent (int): The exponent of the entire multinomial, default value is 1.

        powersum (list): The integer value that all term exponents in the multinomial should
          sum to (or be less than).
    
        possible_powers (list): The possible powers based on the exponent in the multinomial.
    """
    def __init__(self, power, coeff, arrowings, exponent=1):
        """Sets up the multinomial.
        
        Args:
            powers (list): The power on each of the *unexpanded* variables in the multinomial;
              of the form (x^2+y^2+z^2) => 2.

            coeff (list): The coefficient on each of the variables in the multinomial; e.g.
              (x^2 + a y^2) => a. For more than two variables, each color that *has* arrowing
              gets the coefficient while the others get 1.

            exponent (int, optional): The exponent of the entire multinomial, default is 1.
            arrowings (list): The exponents of the arrow terms.
        """
        self.power = power
        self.coeff = coeff
        self.arrowings = arrowings
        self.exponent = exponent
        self.powersum = power*exponent
        """Returns the integer value that all term exponents in the multinomial should
        sum to (or be less than)."""
        self.possible_powers = list(range(0,power*exponent+1, power))
        """For each variable being considered, determines the possible powers based
        on the exponent in the multinomial."""

    def __str__(self):
        #We want to print the multinomial out in a nice, readable way, similar to how
        #they are presented in Mathematica.
        contents = ' + '.join(["{}^{}".format(self.coeff if a else 1, self.power) for a in self.arrowings])
        return "({})^{}".format(contents, self.exponent)

    def normed_seq(self, seq):
        """Normalizes the specified sequence using the powers of unexpanded terms in 
        the multinomial.
        
        Args:
            seq (list): A list of exponents in an *expanded* term.

        Returns:
            norm_seq (list): The normalized lust of exponents.
        """
        return [int(ai/self.power) for ai in seq]

    def nchoosekm(self, sequence):
        """Returns the number of different ways to partition an n-element
        set into disjoint subsets of sizes k1, ..., km.

        Args:
            sequence (tuple): An un-normed tuple of form (k1, k2, k3).
        
        Returns:
            nckm (int): The value of the multinomial coefficient for the series.
        """
        prod = 1
        if not all(seq%self.power == 0 for seq in sequence):
            return 0
        else:
            normseq = self.normed_seq(sequence)
            len_seq = len(sequence)
            for i in range(len_seq):
                nsum = sum(normseq[0:i+1])
                prod *= Multinomial.nchoosek(nsum, normseq[i])

            #Add the contribution from the coefficients of the variable *inside*
            #the multionomial.
            pcoeff = 1
            for iseq, arrow in enumerate(self.arrowings):
                if arrow:
                    pcoeff *= self.coeff**normseq[iseq]

            return prod*pcoeff
        
    @staticmethod
    def nchoosek(n, k):
        """This implementation was taken from "Binomial Coefﬁcient Computation: Recursion 
        or Iteration?" by Yannis Manolopoulos, ACM SIGCSE Bulletin InRoads, Vol.34, No.4, 
        December 2002. http://delab.csd.auth.gr/papers/SBI02m.pdf It is supposed to be robust 
        against large, intermediate values and to have optimal complexity.

        Args:
            n (int): The value of n in n choose k.
            k (int): The value of k in n choose k.

        Return:
            t (int): The value of n choose k.
        """
        if not (-1 < k < n+1): #pragma: no cover
            return 0
        if k==0 and n == 0:
            return 1
        t = 1
        if k < n-k:
            for i in range(n, n-k, -1):
                t = t*i/(n-i+1)
        else:
            for i in range(n, k, -1):
                t = t*i/(n-i+1)

        return t

def group(gen): #pragma: no cover
    """Generates an entire group using the specified generators by applying generators
    to each other recursively.

    Args:
        gen (list): A list of generators as integers.

    Returns:
        groupi (list): A list of the symmetry group operators.
    """
    from operator import itemgetter as iget
    def g_apply(operations, source, groupi=None):
        """Applies the specified group operations to the source list of elements and then
        appends it to the group if it is unique.

        Args:
            operations (list): The symmetry group operation.
            source (list): Object for group to act on.
            groupi (list, optional): The growing symmetry group. Default is None.

        Returns:
            result (list): The result of the symmetry group on the object.
        """
        result = list(iget(*operations)(source))
        if groupi is not None and result not in groupi:
            groupi.append(result)
        return result

    #Make sure the group is zero-based for python.
    if not 0 in gen[0]:
        ngens = [list([e-1 for e in g]) for g in gen]
    else:
        ngens = gen

    groupi = []
    for i in ngens:
        for j in ngens: #filter(lambda k: k!=i, ngens):
            c = g_apply(i, j, groupi)
            d = g_apply(i, c, groupi)
            while d != c:
                d = g_apply(i, d, groupi)

    while True:
        group2 = []
        for i in ngens:
            for h in groupi:
                d = g_apply(i, h)
                if d not in groupi and d not in group2:
                    group2.append(d)

        groupi.extend(group2)
        if len(group2) == 0:
            break
    return(groupi)

def _group_to_cyclic(group, limit=None):
    """Determines the degeneracy of each r-cycle in the specified group operations.

    Args:
        group (list): The a list of the symmetry group operations.
        limit (list, optional): The the start and finishing indices of the desired group.
          Deafult None.

    Returns:
        result (list): The group converted to a cyclic group.
    """
    result = []
    #We allow filtering so that the unit testing can access the cyclic form of the group.
    if 0 not in group[0][0]: #pragma: no cover
        group = [[[j - 1 for j in i] for i in t] for t in group]
    
    if limit is not None: #pragma: no cover
        filtered = group[limit[0]:limit[1]]
    else:
        filtered = group

    for operation in filtered:
        #visitedp has the same # of elements as the site group operation and
        #is used to make sure each element in the array is visited as
        #we loop through in a *non-sequential* order.
        #visitedc has the same number of elements as the arrow group
        #opertaion and is used to make sure each element of that array
        #is visited as we loop through it in a "non-sequential" order.
        visitedp = [0]*len(operation[0])
        visitedc = [0]*len(operation[1])
        polynomials = {}
        
        arrow_cycle_len = []
        #We start by finding the cylce lengths of the arrow group
        #operations so that we can later see how many of them are
        #divosors of the site group operations.
        while 0 in visitedc:
            #Start with the first elemenet in the arrow group that
            #hasn't been visited yet. All cycles have length > 0.
            cursor = vindex = visitedc.index(0)
            cyc_len = 1
            visitedc[cursor] = 1
            cursor = operation[1][cursor]
            while cursor != vindex:
                visitedc[int(cursor)] = 1
                cyc_len += 1
                cursor = int(operation[1][int(cursor)])
            arrow_cycle_len.append(cyc_len)

        while 0 in visitedp:
            #Start with the first element in the site group that hasn't
            #been visited yet. The first non-trivial polynomials have
            #powers > 0.
            cursor = vindex = visitedp.index(0)
            powers = 1 
            #change the current position to having been visited; move the cursor.
            visitedp[cursor] = 1
            cursor = operation[0][cursor]
            #The power of the variables in the polynomials is equal to
            #the number of group operations separating the cursor's
            #current position from its *value* in the group operations
            #list.
            while cursor != vindex:
                visitedp[cursor] = 1
                powers += 1
                cursor = operation[0][cursor]
            #We now have everything need to construct part of the
            #polynomial. This is done by taking powers and using it to
            #construct an array of length equal to the number of
            #elements in the system each entry in the array is set to
            #be equal to powers.

            #To get the coefficients right we need to see how many of
            #the arrow cycles are divisors in length of the current
            #cycle.
            coef = sum([l for l in arrow_cycle_len if powers%l == 0])
            polykey = (powers, coef)
            if polykey not in polynomials:
                polynomials[polykey] = 1
            else:
                polynomials[polykey] += 1
        result.append(polynomials)
        
    return result

def polya(concentrations, group, arrowings=None, debug=False):
    """Uses a group and concentrations to find the number of unique arrangements as described by 
    polya.

    Args:
        concentrations (list):: A list of integers specifying how many of each coloring should
          be present in each of the enumerated lists.

        group (list): A list of group operations for permuting the colorings.
        arrowings (int, optional): The number of arrows present. Default is None.
        debug (bool, optional): True if the code is being debugged.

    Returns:
        polya (int): The polya number for the system.

    Raises:
        ValueError: if the concentrations don't sum to the size of the group operation.
    """
    import itertools
        
    if arrowings is None:
        arrowings = [False]*len(concentrations)
    elif isinstance(arrowings, int):
        arrowings = [False]*(len(concentrations)-arrowings) + [True]*arrowings
        
    #This is to check that the concentrations sum to the number of sites the group is
    #operating on
    if sum(concentrations) != len(group[0][0]):
        raise ValueError("The concentrations don't sum to the number of things the group is acting on!")
            
    len_g = len(group)
    for k, g in itertools.product(range(2),range(len_g)):
        if 0 not in group[g][k]:
            group[g][k] = [j-1 for j in group[g][k]]

    polyndict = {}
    #The operations in the group are used to construct the unique polynomials for each operation.
    for polynomials in _group_to_cyclic(group):
        #Construct a product of multinomials for this group operation.
        p = Product(1,concentrations)
        for exp, coeff in polynomials:
            p.multinoms.append(Multinomial(exp, coeff, arrowings, polynomials[(exp, coeff)]))

        key = str(p)
        if key not in polyndict:
            polyndict[key] = p
        else:
            polyndict[key].coefficient += 1

    if debug: #pragma: no cover
        for key in polyndict:
            print((str(polyndict[key]), " => ", polyndict[key].coeff()))
            
    pp = 0
    for p in list(polyndict.values()):
        pp += p.coeff()
        
    rad = sum([p.coeff() for p in list(polyndict.values())])
    return int(rad/float(len(group)))



"""Contains routines used to map the enumerated structures to real space."""

def map_enumStr_to_real_space(system_data,structure_data,minkowskiReduce):
    """Maps an enumerated structure back to real space. Returns a
    dictionary containing the real space data.

    Args:
        system_data (dict): A dictionary containing all the information about the sysytem with
            keys:

                "plattice": The parrent lattice vectors.

                "dvecs": The atomic basis vectors.

                "title": The name of the system.

                "bulksurf": 'bulk' or 'surface'.

                "nD": The number of atomic basis vectors.

                "k": The number of atomic species.

                "eps": Finite precision tolerance.

        sturture_data (dict): A dictionary containing the information for this structure with
            keys:

                "strN": The structure number.

                "hnfN": The HNFs number.

                "hnf_degen": The HNFs degeneracy.

                "lab_degen": The label degeneracy.

                "tot_degen": The total degeneracy for the structure.

                "sizeN": The system size.

                "n": Number of structure within this size.

                "pgOps": The number of point group operations.

                "diag": The diagonal of the SNF.

                "HNF": The HNF matrix.

                "L": The left transform matrix.

                "labeling": The atomic labeling for the structure.

                "directions": The arrow labeling for the structure.

        minkowskiReduce (bool): Logical indicating if basis should be reduced.

    Returns:
        space_data (dict): A dictionary of the structure mapped to real space with keys:

            "sLV": The super lattice vectors as rows in a matrix.

            "aBas": The attomic basis vectors.

            "spin": A list of the occupanices.

            "gIndx": The integer group index.

            "x": A list of the concentrations of each atom type.
    """
    from numpy import matmul, allclose, matrix
    
    n_d = system_data["nD"]
    n = structure_data["n"]

    # DEFINE the non-zero elements of the HNF matrix
    a = structure_data["HNF"][0][0]
    b = structure_data["HNF"][1][0]
    c = structure_data["HNF"][1][1]
    d = structure_data["HNF"][2][0]
    e = structure_data["HNF"][2][1]
    f = structure_data["HNF"][2][2]

    p_bas = system_data["dvecs"]
    S = structure_data["diag"]
    plv = system_data["plattice"]
    HNF = structure_data["HNF"]
    eps = system_data["eps"]
    L = structure_data["L"]
    # Compute the superlattice vectors
    slv = matmul(plv,HNF).tolist()
    # Find the coordinates of the basis atoms
    g_indx = []

    if minkowskiReduce:
        slv = list(map(list,zip(*_minkowski_reduce_basis(list(map(list,zip(*slv))),eps))))
        
    # Find each atomic position from the g-space information
    a_bas = []
    ic = 0  # Keep track of the number of points mapped so far
    # Loop over the number at sites/parent cell (the d set)it
    for iD in range(0, n_d):  
        # For the limits on the loops, see the "interior_points.pdf" write-up
        for z1 in range(a):
            for z2 in range(int((b*z1)/a), int(c+(b*z1)/a)):
                for z3 in range(int(z1*(d-(e*b)/c)/a+(e*z2)/c), int(f+z1*(d-(e*b)/c)/a+(e*z2)/c)):
                    ic +=1
                    # if ic > n: #pragma: no cover
                    #     from .msg import err
                    #     err("Problem with basis atoms in map_enpStr_to_real_space...")
                    #     exit()
                    # Atomic basis vector in Cartesian coordinates
                    temp = matmul(plv,[z1,z2,z3]).tolist()
                    temp2 = [temp[i]+p_bas[iD][i] for i in range(len(p_bas[iD]))]
                    a_bas.append(temp2)
                    # Map position into the group
                    greal = matmul(L,[float(z1),float(z2),float(z3)]).tolist() 
                    g = [int(i) for i in greal] # Convert the g-vector from real to integer
                    if not allclose(greal,g,rtol=eps,atol=eps): #pragma: no cover
                        from .msg import err
                        err("map2G didn't work in map_enumStr_to_real_space")
                        exit()
                    # Bring the g-vector back into the first tile
                    g = [g[i]%S[i] for i in range(len(S))] 
                    # g_indx is the index in the configuration string that
                    # tells us which atom type is used at this position

                    g_indx.append((iD)*S[0]*S[1]*S[2]+g[0]*S[1]*S[2]+g[1]*S[2]+g[2])

    if ic != n*n_d: #pragma: no cover
        from .msg import err
        err("ERROR: map_enumStr_to_real_space: Didn't find the correct # of basis atoms")
        exit()

    k = system_data["k"]
    x = []
    for i in range(k):
        x.append(0.0)
        
    labeling = structure_data["labeling"]
    spin = []
    if k % 2 == 0:
        for iAt in range(0, n*n_d):
            i = int(labeling[g_indx[iAt]])
            digit = i-k//2 # convert 0..k-1 label to spin variable -k/2..k/2
            x[i] += 1  # Keep track of the concentration of each atom type
            if digit < 0:
                spin.append(digit)
            else:
                spin.append(digit+1) # skip 0 as a spin if k is even
    else:
        for iAt in range(0, n*n_d):
            i = int(labeling[g_indx[iAt]])
            spin.append(i-k//2)
            x[i] += 1 # Keep track of the concentration of each atom type

    x = [i/float(n*n_d) for i in x]

    space_data = {"sLV": list(map(list,zip(*slv))), "aBas": a_bas, "spin": spin, "gIndx": g_indx, "x": x}

    return space_data

def _minkowski_reduce_basis(IN,eps):
    """Performs a minkowski reduction on the basis atoms.

    Args:
        IN (list): The input basis as rows of a matrix.
        eps (float): Floating point tolerance.

    Returns:
        OUT (list): The reduced basis vectors as rows of a matrix.

    Raises:
        ValueError: if the input vectors are linearly dependent.
    """

    from numpy import allclose, linalg, array, matrix
    from copy import deepcopy

    limit = 10

    if allclose(linalg.det(IN),0.0,rtol=eps,atol=eps):
        raise ValueError("Input basis for 'minkowski_reduce_basis' was not linearly independent")

    OUT = deepcopy(IN)

    # Keep applying the greedy algorithm until the vectors come out already sorted
    for it in range(1, limit +1):
        # Sort the three vectors into ascending order
        temp = deepcopy(OUT)
        norms = linalg.norm(temp,axis=1).tolist()
        tt = list(range(3))
        tt.reverse()
        for i in tt:
            idx = norms.index(max(norms))
            temp[i] = OUT[idx]
            norms[idx] = 0
        OUT = deepcopy(temp) # Copy the sorted vectors back to OUT
        (OUT[0], OUT[1], OUT[2]) = _reduce_C_in_ABC(OUT[0],OUT[1],OUT[2],eps)
        if linalg.norm(OUT[2]) >= (linalg.norm(OUT[1])-eps):
            break

    if not _minkowski_conditions_check(OUT,eps): #pragma: no cover
        from .msg import err
        err("ERROR in minkowski_reduce_basis: Minkowski conditions not met."
            "Number of iterations: {}".format(str(limit)))
        exit()
    
    # we want to make sure that the det is positive.
    # NOTE: This *destroys* the mathematical picture of a "greedy reduced basis" (Minkowski), but
    #       from a physical point of view we don't care ;-)
    #       Either way, the basis is as orthogonal as possible.
    if linalg.det(OUT) < 0:
        temp[0] = OUT[1]
        OUT[1] = OUT[2]
        OUT[2] = temp[0]

    return OUT

def _minkowski_conditions_check(basis,eps):
    """This function checks the minkowski conditions for a 3D lattice
    basis.

    Args:
        basis (list): The atomic basis vectors.
        eps (float): Finitie precision tolerance.

    Returns:
        minkowski_check (bool): True if minkowski conditions are met.
    """
    from numpy import linalg
    from .msg import err
    
    b1 = basis[0]
    b2 = basis[1]
    b3 = basis[2]

    minkowski_check = True
    if linalg.norm(b1) > (linalg.norm(b2) + eps):
        minkowski_check = False
        err("Minkowski_condition 1 failed: b1 > b2")

    if linalg.norm(b2) > (linalg.norm(b3) + eps):
        minkowski_check = False
        err("Minkowski_condition 2 failed: b2 > b3")

    if linalg.norm(b2) > (linalg.norm([b1[i]+b2[i] for i in range(len(b1))])+eps):
        minkowski_check = False
        err("Minkowski_condition 3 failed: b2 > b1+b2")

    if linalg.norm(b2) > (linalg.norm([b1[i]-b2[i] for i in range(len(b1))])+eps):
        minkowski_check = False
        err("Minkowski_condition 4 failed: b2 > b1-b2")

    if linalg.norm(b3) > (linalg.norm([b1[i]+b3[i] for i in range(len(b1))])+eps):
        minkowski_check = False
        err("Minkowski_condition 5 failed: b3 > b1+b3")

    if linalg.norm(b3) > (linalg.norm([b3[i]-b1[i] for i in range(len(b1))])+eps):
        minkowski_check = False
        err("Minkowski_condition 6 failed: b3 > b3-b1")

    if linalg.norm(b3) > (linalg.norm([b2[i]+b3[i] for i in range(len(b2))])+eps):
        minkowski_check = False
        err("Minkowski_condition 7 failed: b3 > b2+b3")

    if linalg.norm(b3) > (linalg.norm([b3[i]-b2[i] for i in range(len(b2))])+eps):
        minkowski_check = False
        err("Minkowski_condition 8 failed: b3 > b3-b2")

    if linalg.norm(b3) > (linalg.norm([b1[i]+b2[i]+b3[i] for i in range(len(b2))])+eps):
        minkowski_check = False
        err("Minkowski_condition 9 failed: b3 > b1+b2+b3")

    if linalg.norm(b3) > (linalg.norm([b1[i]-b2[i]+b3[i] for i in range(len(b2))])+eps):
        minkowski_check = False
        err("Minkowski_condition 10 failed: b3 > b1-b2+b3")

    if linalg.norm(b3) > (linalg.norm([b1[i]+b2[i]-b3[i] for i in range(len(b2))])+eps):
        minkowski_check = False
        err("Minkowski_condition 11 failed: b3 > b1+b2-b3")

    if linalg.norm(b3) > (linalg.norm([b1[i]-b2[i]-b3[i] for i in range(len(b2))])+eps):
        minkowski_check = False
        err("Minkowski_condition 12 failed: b3 > b1-b2-b3")

    return minkowski_check

def _reduce_C_in_ABC(A,B,C,eps):
    """This routine takes three vectors, A,B,C, defining a lattice, and
    reduces the last one so that it is as close as possible to the
    origin while remaining in an affine plane, which is parallel to
    the A-B plane but which passes through the end of the C
    vector. See Lecture notes in computer science, ISSN 0302-974, ANTS
    - VI : algorithmic number theory, 2004, vol. 3076, pp. 338-357
    ISBN 3-540-22156-5

    Args:
        A (list): The first vector.
        B (list): The second vector.
        C (list): The vector to be reduced.
        eps (float): Finite precision tolerance.

    Returns:
        A (list): The vector A.
        B (list): The vector B.
        C (list): The reduced vector C.
    """
    from numpy import cross, linalg, dot, allclose, matmul, array
    from copy import deepcopy
    from math import floor
    
    old_abc = deepcopy([A,B,C])
    # Use Gaussian reduction to reduce the A,B 2D basis so that it is
    # itself Minkowski reduced. If this is done, then the closest
    # lattice point (in A,B plane) to the projection of C (into the
    # A,B plane) is guaranteed to be one of the corners of the unit
    # cell enclosing the projection of C
    (A,B) = _gaussian_reduce_two_vectors(A,B,eps)

    # First thing to do is find the (real, not lattice) point in the
    # affine plane A,B + C that is nearest the origin. Call this T.
    cpd_ab = [i/linalg.norm(cross(A,B)) for i in cross(A,B)]
    T = [C[i] - cpd_ab[i]*dot(C,cpd_ab) for i in range(3)]

    if not allclose(dot(T,cross(A,B)),0,atol=eps,rtol=eps): #pragma: no cover
        from .msg import err
        err("{0} Projection of C into A,B plane failed".format(str(dot(T,cross(A,B)))))

    # Now find the four points of the A,B lattice, in the affine
    # plane, that enclose the point T
    abc = [A,B,C]
    abcinv = linalg.inv(abc)

    LC = [int(floor(i +eps)) for i in matmul(T,abcinv).tolist()]
    # Compute the distance from T to each of the four corners of the cell and pick
    # the one that is the closest.
    corners = array([[0,0,0],[1,0,0],[0,1,0],[1,1,0]])
    dist = []
    for i in range(0,4):
        temp1 = corners[i] + array(LC)
        temp2 = array(T) -matmul((corners[i] + array(LC)),abc)
        dist.append(linalg.norm(array(T) -matmul((corners[i] + array(LC)),abc)))

    idx = dist.index(min(dist))

    if idx == 0:
        temp1 = [corners[0][i] + LC[i] for i in range(3)]
        temp2 = matmul(temp1,abc).tolist()
        C = [C[i] - temp2[i] for i in range(len(C))]
    elif idx == 1:
        temp1 = [corners[1][i] + LC[i] for i in range(3)]
        temp2 = matmul(temp1,abc).tolist()
        C = [C[i] - temp2[i] for i in range(len(C))]
    elif idx == 2:
        temp1 = [corners[2][i] + LC[i] for i in range(3)]
        temp2 = matmul(temp1,abc).tolist()
        C = [C[i] - temp2[i] for i in range(len(C))]
    elif idx == 3:
        temp1 = [corners[3][i] + LC[i] for i in range(3)]
        temp2 = matmul(temp1,abc).tolist()
        C = [C[i] - temp2[i] for i in range(len(C))]
    else: #pragma: no cover
        from .msg import err 
        err("Case failed in reduce_C_in_ABC Lattice coordinates in the A,B plane: {0}".format(' '.join([str(i) for i in LC])))

    abc = [A,B,C]
    abcinv = linalg.inv(abc)
    temp = matmul(list(map(list,zip(*abcinv))),list(map(list,zip(*old_abc)))).tolist()
    for i in range(3):
        for j in range(3):
            if abs(temp[i][j] - int(round(temp[i][j]))) > eps: #pragma: no cover
                from .msg import err
                err("Lattice was not preserved in reduce_C_in_ABC")
                exit()

    return A, B, C

def _gaussian_reduce_two_vectors(U,V,eps):
    """This routine takes two vectors (in three-space) and reduces them to
    form a shortest set (Minkowski reduced). The idea is to subtract
    multiples of U from V so that the new V is as close to the origin
    as any lattice point along the line that passes through U in the
    direction of V. The process is repeated until the new vector isn't
    shorter than the other. It's pretty obvious if you do an example
    by hand. Also see 3.1 of Lecture notes in computer science, ISSN
    0302-974, ANTS - VI: algorithmic number theory, 2004, vol. 3076,
    pp. 338-357 ISBN 3-540-22156-5. Fixes made Apr 2012 GLWH (not sure
    if they made a practical difference though)

    Args:
        U (list): A vector.
        V (list): Another vector.
        eps (float): Finite precision tolerance.

    Returns:
        U (list): The reduced vector U.
        V (list): The reduced vector V.
    """

    from numpy.linalg import norm
    from numpy import dot

    it = 0
    if norm(U) > (norm(V) - eps):
       # Make sure that the {U,V} are listed in ascending order; ||U||<||V||
       temp = U
       U = V
       V = temp # Keep V as the longest vector

    done = False
    it = 1
    while not done:
        if it > 10: # pragma: no cover
            from .msg import err
            err("gaussian_reduce_two_vectors failed to converge in 10 iterations")
            exit()
        R = [V[i]-int(round(dot(U,V)/float(dot(U,U))+1E-10))*U[i] for i in range(3)] #Shorten V as much as possible
        V = U # Swap U and V (so U remains the shortest)
        U = R
        if norm(U) >= (norm(V) - eps):
            done = True
        it += 1

    # Make sure that the {U,V} are listed in ascending order on exit; ||U||<||V||
    temp = U
    U = V
    V = temp

    return U, V

def cartesian2direct(sLV,aBas, eps):
    """This routine takes three lattice vectors and a list of atomic basis
    vector in Cartesian coordinates and converts them to direct
    ("lattice") coordinates.

    Args:
        sLV (list): The superlattice vectors in cartesian coordinates as rows of a matrix.
        aBas (list): Atomic basis vectors in cartesian coordinates.
        eps (float): Finite precision tolerance.

    Returns:
        aBas (list): The atomic basis vectors in direct coordinates.
    """
    from numpy import linalg, matmul, array
    
    n_at = len(aBas)
    slv_inv = linalg.inv(sLV)
    # Convert aBas to DIRECT COORDINATES
    for iAt in range(n_at):
        aBas[iAt] = matmul(aBas[iAt],slv_inv) # Put positions into
        # "direct" coordinates This keeps the atomic coordinates inside
        # the first unit cell--- not necessary but aesthetically
        # pleasing.

        while any(aBas[iAt] >= (1.0-eps)) or any(aBas[iAt] < (0.0-eps)):
            aBas[iAt] = array([i if i < (1.0-eps) else i-1.0 for i in aBas[iAt]])
            aBas[iAt] = array([i if i >= (0.0-eps) else i+1.0 for i in aBas[iAt]])

    return aBas



import numpy as np

# The dictionary of all the elements on the periodic table
element_volume ={"H":37.2958,"He":32.1789,"Li":21.2543,"Be":8.49323,"B":7.24205,"C":5.68741,
                 "N":46.6002,"O":22.2802,"F":17.0258,"Ne":21.7346,"Na":23.2596,"Mg":23.3928,
                 "Al":16.6075,"Si":7.8511,"P":9.1459,"S":17.1672,"Cl":35.2074,"Ar":36.3829,
                 "K":71.5278,"Ca":43.4353,"Sc":25.6478,"Ti":18.1565,"V":13.7718,"Cr":11.9439,
                 "Mn":19.3207,"Fe":11.82,"Co":11.1838,"Ni":10.9036,"Cu":11.7615,"Zn":13.311,
                 "Ga":18.4496,"Ge":19.3638,"As":20.4270,"Se":58.6173,"Br":33.3170,"Kr":46.7873,
                 "Rb":87.3384,"Sr":56.1889,"Y":33.0792,"Zr":23.8327,"Nb":17.9685,"Mo":15.6279,
                 "Tc":14.5458,"Ru":13.9206,"Rh":13.718,"Pd":14.716,"Ag":17.1045,"Cd":18.7161,
                 "In":26.6861,"Sn":29.3238,"Sb":27.1733,"Te":62.3227,"I":24.3807,"Xe":59.582,
                 "Cs":110.723,"Ba":63.253,"Hf":23.1748,"Ta":18.1323,"W":15.7772,"Re":14.8694,
                 "Os":14.5485,"Ir":14.1558,"Pt":15.0591,"Au":16.9793,"Hg":27.6914,"Tl":29.2949,
                 "Pd":30.3218,"Bi":31.2849}

def get_lat_param_element(lat_vecs,n_basis_atoms,element):
    """Finds the lattice parameter of an element for the given set of lattice.

    Args:
        lat_vecs (list): The lattice vectors for the system.
        n_basis_atoms (int): The number of atoms in the atomic basis.
        element (str): The symbol for the element.
    
    Returns:
        lat_param (float): The lattice parameter.
    """

    lat_vol = abs(np.linalg.det(lat_vecs)/n_basis_atoms)
    atom_vol = element_volume[element]/lat_vol

    return atom_vol**(1./3.)   

def get_lattice_parameter(elements, concentrations, lat_vecs, n_basis_atoms, default_title,remove_zeros=False):
    """Finds the lattice parameters for the provided atomic species using Vagars law.

    Args:
        elements (list of str): A list of the elements in the system.
        default_title (str): The default system title.
        concentrations (list of int): The concentrations of each element.
        lat_vecs (list): The lattice vectors for the system.
        n_basis_atoms (int): The number of atoms in the atomic basis.
        remove_zeros (bool): True if zeros are to be removed from the elements list.

    Returns:
        lat_param (float): The lattice parameter of the system.

    Raises:
        ValueError: if the number of elements doesn't match the len of the concentration array.
    """

    if elements is None:
        lat_param = 1.0
        title = default_title
    else:
        if len(elements) != len(concentrations):
            raise ValueError("You have provided {0} element names when {1} elements are present "
                "in the system. Please provide the correct number of elements."
                .format(len(elements),len(concentrations)))

        else:
            title = ""
            lat_param = 0
            for i, elem in enumerate(elements):
                lat_param += concentrations[i]*get_lat_param_element(lat_vecs,n_basis_atoms,elem)
                if concentrations[i] > 0:
                    title += " {0} ".format(elem)
                elif not remove_zeros:
                    title += " {0} ".format(elem)
            lat_param = float(lat_param) / sum(concentrations)
            title = "{0} {1}\n".format(title,default_title.strip())
    return lat_param, title



"""Methods for reading and writing enumeration and polya-counting results."""
def read_group(fname):
    """Reads the symmetry group in from the 'rot_perms' styled group
    output by enum.x.
    Args:
        fname (str): Path to the file to read the group from.
    Returns:
        group (numpy ndarray): The symmetry group.
    """
    i=0
    groupi = []
    with open(fname) as f:
        for line in f:
            if i > 5:
                if ('Perm #:') in line:
                    groupi.append(list(map(int, line.split()[4::])))
                else:
                    groupi[-1] += list(map(int, line.split()))
            i += 1
    from numpy import array
    return(list(map(list, array(groupi)-1)))

def read_lattice(filename='lattice.in'):
    """Reads the lattice.in file.
    Args:
        filename (str, optional): The filename to be read in. Default is 'lattice.in'.
    Returns: 
        result (dict): A dictionary with the following fields:
            "sizes": the range of cell sizes,
            "lat_vecs": lattice vectors of the parent cell,
            "nspecies": the number of atomic species in the enumeration,
            "basis_vecs": basis vectors of the parent cell,
            "is_crestricted": logical that indicates if the concentrations will be restricted,
            "arrows": logical that indicates if arrows are present,
            "concs": array of the concentrations in format [1,3].
    """
    from .msg import info
    with open(filename,'r') as f:
        lines = [l for l in f.readlines() if l.strip()[0] != '#']
        
    sizes = [int(el) for el in lines[0].split(' ')]
    bulk = lines[1].strip()[0].lower() == 'b'
    lat_vecs = [list(map(float, l.split())) for l in lines[2:5]]
    nspecies = int(lines[5])
    nbas = int(lines[6])
    bas_vecs = [list(map(float, l.split())) for l in lines[7:7+nbas]]

    arrows = False
    res_conc = False
    if lines[7+nbas].strip() == 'F':
        info('No concentration restrictions given.')
        concs = []
    else:
        res_conc = True
        info('Concentrations are being restricted.')
        arrows = lines[7+nbas+1].strip()
        if arrows == 'F':
            arrows = []
            info('No displacement directions are included')
        else:
            arrows = True
        concs = [list(map(float, l.split())) for l in lines[7+nbas+2:]]

    result = {
        "bulk": bulk,
        "sizes": sizes,
        "lat_vecs": lat_vecs,
        "nspecies": nspecies,
        "basis_vecs": bas_vecs,
        "is_crestricted": res_conc,
        "arrows": arrows,
        "concs": concs
    }
    return result

# This method reads in the enum.in file.
def read_enum(filename="enum.in"):
    """Reads the list of structures and the number of random candidates to enumerate
    from the 'enum.in' styled file. It should contain only integers and have:
       HNF concs num_wanted
    where,
    HNF: the list of 6 independent entries in the HNF matrix;
    concs: list of integer species concentrations;
    num_wanted: number of random structures to draw from the enumerated list.
    Args:
        filename (str, optional): The filename and path. Default 'enum.in'.
    
    Returns:
        systems (list of lists): A list containing the lists of [HNF, concs, num_wanted].
    """    
    from numpy import loadtxt
    raw = loadtxt(filename, int, ndmin=2)
    systems = []
    for i in raw:
        HNF = i[0:6]
        sys_conc = i[6:-1]
        num_wanted = i[-1]
        systems.append((HNF, sys_conc, num_wanted))
    return systems

def write_enum(params, outfile="enum.out"):
    """Writes a 'struct_enum.out' style preamble to the specified output file
    using the enumeration parameters gleaned from a 'lattice.in' styled file.
    Args:
        params (dict): values returned from method:read_lattice().
        outfile (str, optional): path to desired output file. Default is 'enum.out'.
    """
    from datetime import datetime
    lines = []
    lines.append("Random structure enumeration: {0}".format(str(datetime.now())))
    lines.append("bulk" if params["bulk"] else "surface")
    lines.extend([" {0:.7f}       {1:.7f}       {2:.7f}".format(*l) for l in params["lat_vecs"]])
    lines.append("    {0:d}".format(len(params["basis_vecs"])))
    lines.extend(["  {0:.7f}       {1:.7f}       {2:.7f}".format(*l) for l in params["basis_vecs"]])
    lines.append(" {0:d}-nary case".format(params["nspecies"]))
    lines.append("   " + '   '.join(map(str, params["sizes"])))
    lines.append(" 1e-7")
    lines.append("Concentration check:")
    lines.append("    T" if params["is_crestricted"] else "    F")
    lines.extend(["full list of labelings (including incomplete labelings) is used ",
                  "Equivalency list: 1 #Not used in the random enumeration algorithm ",
                  "start   #tot      HNF     Hdegn   labdegn   Totdegn   #size idx    "
                  "pg    SNF             HNF                 Left "
                  "transform                          labeling                 "
                  "directions",
                  ""])
        
    with open(outfile, 'w') as f:
        f.write('\n'.join(lines))

def write_struct_enum(params):
    """Writes a 'struct_enum.in' file for the executable enum.x.
    Args:
        params (dict): values returned from method:read_lattice().
    """

    with open("struct_enum.in","w+") as struct_file:
        struct_file.write("System \n")

        if params["bulk"] == True:
            struct_file.write("bulk \n")
        else:
            struct_file.write("surface \n")

        latt = ""
        i = 0
        for vec in params["lat_vecs"]:
            for point in vec:
                latt += str(point) + " "
            if i < 2:
                latt += "\n"
            i += 1

        struct_file.write(latt + " \n")

        struct_file.write(str(params["nspecies"]) + " \n")

        occupancy = ""
        for i in range(params["nspecies"]):
            if i < params["nspecies"]-1:
                occupancy += str(i) +"/"
            else:
                occupancy += str(i)
    
        basis = ""
        for vec in params["basis_vecs"]:
            basis += "{}  {} \n".format(" ".join([str(point) for point in vec]),occupancy)
        
        struct_file.write(str(len(params["basis_vecs"])) + " \n")
        struct_file.write(basis + " \n")

        size_range = ""
        i = 0
        for point in params["sizes"]:
            size_range += str(point) + " "

        struct_file.write(size_range + " \n")
        struct_file.write("0.00001 \n")
        struct_file.write("full \n")

def which(program):
    """Determines if and where an executable exists on the users path.
    This code was contributed by Jay at http://stackoverflow.com/a/377028
    Args:
        program (str): The name, or path for the program.
    Returns:
        The program or executable.
    """
    import os
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath and is_exe(program):
        return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def create_labeling(config):
    """This routine takes a string of atomic locations as a vector and the
    atomic concentrations and returns a unique labeling.
    Args:
        config (list of int): A list of integers describing the arrangement of atoms on
          the lattice.
    Returns:
        label (str): The atomic labeling for the arrangement.
        arrow (str): The arrow directions for the arrangement.
    """
    label = ''
    arrow =  ''
    if isinstance(config[0],list):
        for i in config:
            label += str(i[1]-1)
            arrow += str(i[0]+1)
    else:
        for i in config:
            label += str(i-1)
            arrow += '0'
    
    return(label,arrow)

def read_enum_out(args):
    """Reads the enum.out file and builds a dictionary with the needed
    information to construct a POSCAR.
    
    Args:
        args (dict): The makeStr.py input arguments.
    Returns:
        system (dict): A dictionary with the keys:
            "plattice": The parrent lattice vectors.
            "dvecs": The atomic basis vectors.
            "title": The name of the system.
            "bulksurf": 'bulk' or 'surface'.
            "nD": The number of atomic basis vectors.
            "k": The number of atomic species.
            "eps": Finite precision tolerance.
        structure_data (list of dict): A list of dictionaries for each desired structure. Each
            dictionary contains:
                "strN": The structure number.
                "hnfN": The HNFs number.
                "hnf_degen": The HNFs degeneracy.
                "lab_degen": The label degeneracy.
                "tot_degen": The total degeneracy for the structure.
                "sizeN": The system size.
                "n": Number of structure within this size.
                "pgOps": The number of point group operations.
                "diag": The diagonal of the SNF.
                "HNF": The HNF matrix.
                "L": The left transform matrix.
                "labeling": The atomic labeling for the structure.
                "directions": The arrow labeling for the structure.
     """
    from numpy import transpose
    
    # which structures are wanted
    if args["structures"] == None:
        with open(args["input"],"r") as f:
            for count, l in enumerate(f):
                pass
        structures = list(range(1,count-13))
    else:
        structures = args["structures"]
    # open the enum.out style file.
    structf = open(args["input"],"r")

    # we'll build a dictionary of the system data and a list of
    # dictionaries for the structures that are wanted.
    structure_data = []
    system = {}
    system["plattice"] = []
    system["dvecs"] = []
    line_count = 1
    system["nD"] = 0
    for line in structf:
        temp = line.rstrip()
        if not temp.startswith("#"):
            if line_count == 1:
                system["title"] = temp
            if line_count == 2:
                system["bulksurf"] = temp
            if line_count in [3,4,5]:
                vec = [float(v) for v in temp.split()]
                system["plattice"].append(vec)
            if line_count == 6:
                system["nD"] = int(temp.rstrip())
            if system["nD"] != 0 and line_count in range(7,7+system["nD"]):
                vec = [float(v) for v in temp.split()]
                system["dvecs"].append(vec)
            if line_count == 7+system["nD"]:
                system["k"] = int(temp.split('-')[0].strip())
            if line_count == 9 + system["nD"]:
                system["eps"] = float(temp.strip())
            if line_count - (14 + system["nD"]) in structures and system["nD"] !=0:
                data = temp.split()
                this_struct = {}
                this_struct["strN"] = int(data[0])
                this_struct["hnfN"] = int(data[1])
                this_struct["hnf_degen"] = int(data[2])
                this_struct["lab_degen"] = int(data[3])
                this_struct["tot_degen"] = int(data[4])
                this_struct["sizeN"] = int(data[5])
                this_struct["n"] = int(data[6])
                this_struct["pgOps"] = int(data[7])
                this_struct["diag"] = [int(data[8]),int(data[9]),int(data[10])]
                this_struct["HNF"] = [[int(data[11]),0,0],[int(data[12]),int(data[13]),0],
                                      [int(data[14]),int(data[15]),int(data[16])]]
                this_struct["L"] = [[int(data[17]),int(data[18]),int(data[19])],
                                    [int(data[20]),int(data[21]),int(data[22])],
                                      [int(data[23]),int(data[24]),int(data[25])]]
                this_struct["labeling"] = data[26]
                this_struct["directions"] = data[27]                
                structure_data.append(this_struct)
        line_count += 1

    system["plattice"] = transpose(system["plattice"])
        
    return (system, structure_data)

def write_POSCAR(system_data,space_data,structure_data,args):
    """Writes a vasp POSCAR style file for the input structure and system
    data.
    Args:
        system_data (dict): A dictionary of the system_data.
        space_data (dict): A dictionary containing the spacial data
        structure_data (dict):: a dictionary of the data for this structure
        args (dict): Dictionary of user supplied input.
    """

    from numpy import array
    from phenum.element_data import get_lattice_parameter
    from random import uniform
    from itertools import product

    if "{}" in args["outfile"]:
        filename = args["outfile"].format(str(structure_data["strN"]))
    else:
        filename = args["outfile"] + ".{}".format(str(structure_data["strN"]))

    labeling = structure_data["labeling"]            
    g_indx = space_data["gIndx"]
    arrows = structure_data["directions"]

    arrow_directions = [[0,0,0],[0,0,-1],[0,-1,0],[-1,0,0],[1,0,0],[0,1,0],[0,0,1]]
    directions = []

    concs = [0]*system_data["k"]
    for atom in range(structure_data["n"]*system_data["nD"]):
        concs[int(labeling[g_indx[atom]])] += 1
        
    def_title = "{} str #: {}\n".format(str(system_data["title"]),str(structure_data["strN"]))

    lattice_parameter, title = get_lattice_parameter(args["species"],concs,
                                                     system_data["plattice"],system_data["nD"],
                                                     def_title,remove_zeros=args["remove_zeros"])

    for arrow in arrows:
        directions.append(array(arrow_directions[int(arrow)]))
    slv = space_data["sLV"]
    with open(filename,"w+") as poscar:
        poscar.write(title)
        poscar.write("{0:.2f}\n".format(lattice_parameter))
        for i in range(3):
            poscar.write(" {}\n".format(" ".join(
                ["{0: .8f}".format(j) for j in slv[i]])))
        poscar.write("  ")
        # If the species strings were passed in by the user and the
        # user requests there be no zeros in the concentration string
        # then we should remove them from the file. Otherwise we
        # default to leaving them in.
        if (args["remove_zeros"] and args["species"] is not None):
            for ic in concs:
                if ic != 0:
                    poscar.write("{}   ".format(str(ic)))
        else:
            for ic in concs:
                poscar.write("{}   ".format(str(ic)))                    

        poscar.write("\n")
        poscar.write("D\n")
        for ilab, iAt in product(range(system_data["k"]),range(structure_data["n"]*system_data["nD"])):
            rattle = uniform(-args["rattle"],args["rattle"])
            displace = directions[iAt]*args["displace"]*lattice_parameter
            displace += displace*rattle
            if labeling[g_indx[iAt]] == str(ilab):
                out_array = array(space_data["aBas"][iAt]) + displace
                poscar.write(" {}\n".format(
                    "  ".join(["{0: .8f}".format(i) for i in out_array.tolist()])))


def write_config(system_data,space_data,structure_data,args,mapping=None):
    """Writes a MTP config style file for the input structure and system
    data.
    :arg system_data: a dictionary of the system_data
    :arg space_data: a dictionary containing the spacial data
    :arg structure_data: a dictionary of the data for this structure
    :arg args: Dictionary of user supplied input.
    :arg mapping: A dictionary of the species mappings if to be used.
    """

    from phenum.element_data import get_lattice_parameter
    from numpy import array, dot, transpose
    from random import uniform
    from numpy.random import randint
    
    filename = args["outfile"]

    # Get the labeling, group index, structure number and arrow labels
    # from the input data structure.
    labeling = structure_data["labeling"]            
    gIndx = space_data["gIndx"]
    arrows = structure_data["directions"]
    struct_n = structure_data["strN"]

    # The arrow basis.
    arrow_directions = [[1,0,0],[-1,0,0],[0,1,0],[0,-1,0],[0,0,1],[0,0,-1]]
    directions = []

    # Construct the concentrations of the atoms from the labeling by
    # counting the number of each type of atom present in the
    # labeling.
    concs = []
    for i in range(system_data["k"]):
        this_conc = 0
        for atom in range(structure_data["n"]*system_data["nD"]):
            if labeling[gIndx[atom]] == str(i):
                this_conc += 1
        concs.append(this_conc)
    def_title = "{}{}\n".format(str(system_data["title"]),str(structure_data["strN"]))

    # Get the lattice parameter for the atomic species provided by the
    # user.
    if mapping is not None and len(args["species"]) != len(mapping):
        species = [args["species"][i] for i in mapping.values()]
    else:
        species = args["species"]
    
    lattice_parameter, title = get_lattice_parameter(species,concs,
                                                     system_data["plattice"],system_data["nD"],
                                                     def_title,remove_zeros=True)

    # Find out the directions for each arrow.
    for arrow in arrows:
        directions.append(array(arrow_directions[int(arrow)]))

    sLV = list(lattice_parameter*array(space_data["sLV"]))

    # Start writing the data to the file.
    with open(filename,"a+") as poscar:
        # First write the title and the lattice parameter.
        poscar.write("BEGIN_CFG\n Size\n")
        poscar.write("    {}\n".format(structure_data["n"]*system_data["nD"]))
        poscar.write(" SuperCell\n")
        # Then write out the lattice vectors.
        for i in range(3):
            poscar.write("   {}\n".format("      ".join(
                ["{0: .6f}".format(j) for j in sLV[i]])))
        
        poscar.write("  ")

        poscar.write(" AtomData:  id type       cartes_x      cartes_y      cartes_z\n")
        for iAt in range(structure_data["n"]*system_data["nD"]):
            for ilab in range(system_data["k"]):
                rattle = uniform(-args["rattle"],args["rattle"])
                displace = directions[iAt]*args["displace"]*lattice_parameter
                # If the displacement is non zero and we're `rattling`
                # the system then we need to modify the displacement
                # by the amount being rattled.
                displace += displace*rattle
                if labeling[gIndx[iAt]] == str(ilab):
                    # The final atomic position is the position from
                    # the basis plus the total displacement.
                    out_array = list(array(dot(transpose(sLV),space_data["aBas"][iAt])) + displace)
                    if mapping is None:
                        out_lab = ilab
                    else:
                        out_lab = mapping[ilab]
                    poscar.write("             {0}    {1}       {2}\n".format(iAt+1, out_lab, "  ".join(["{0: .8f}".format(i) for i in out_array])))
        poscar.write(" Feature   conf_id  {}\n".format(title.split()[0]))
        poscar.write("END_CFG\n\n")
        



"""Methods for reading and writing enumeration and polya-counting results.

Generates a random subset of possible structures weighted by the
Polya distribution for superstructures.
"""

def _read_struct_enum(): #pragma: no cover
    """Reads in the struct_enum.out file and returns a dictionary with key
    for each cell size, and value a dictionary keyed by HNF with the number
    of unique structures.

    Returns:
        structs (dict): A dictionary containing the structure data for each structure.
    """
    i = 0
    structs = {}
    with open("struct_enum.out") as f:
        for line in f:
            i += 1                        
            if i > 17:
                vals = line.split()
                sn = int(vals[6])
                HNF = tuple(map(int, vals[11:17]))
                concs = tuple([vals[-1].count(c) for c in ["0", "1"]])
                if sn not in structs:
                    structs[sn] = {}
                if HNF not in structs[sn]:
                    structs[sn][HNF] = {}
                    structs[sn][HNF][concs] = 1
                else:
                    if concs in structs[sn][HNF]:
                        structs[sn][HNF][concs] += 1
                    else:
                        structs[sn][HNF][concs] = 1
    return structs

def _distribute(infile, cellsizes, ftype, n=None,seed=None, res_type=None, res_values=None):
    """Returns a dictionary specifying how many of each cell shape, size and concentration
    to enumerate in order to obtain a grand total of 'n' unique cells, distributed according
    to the abundance of unique structures predicted by Polya. See the calling signature for
    method:distribution_fxn(). However, while that method does not guarantee the total 'n',
    this one does; if it is short a few at the end, it tops up the list according to the
    abundances across all cell sizes, shapes, etc.

    Args:
        inlfile (str): The polya.out style input file.
        cellsizes (list): a list of the cell sizes to include in the data set summary.
        ftype (str): one of ['shape', 'conc', 'size', 'all']; determines the number of 
            arguments that the distr. function will accept and average over, thus controlling
            the granularity of the predictions.
        n (int): the number of structure types to return, weighted by their polya distribution.
            If this value is None, then *all* structure types are returned with their unique
            number.
        seed (float): The seed for the random number generator.
    
    Returns:
        result (dict): A dictionary specifying how many of each cell shape, size and
            concentration to enumerate.
    """
    (f, dataset, gtotal) = _distribution(infile, ftype, None, None, cellsizes=cellsizes,
                                         res_values=res_values,res_type=res_type)
    #If they didn't specify an 'n', then we also return all the structures.
    if n is None:
        n = gtotal
    elif n > gtotal:
        from .msg import warn
        warn("The number of unique structures you requested ({}) ".format(n) +
             "exceeds the total number of unique structures ({}). ".format(gtotal) +
             "The code will return *all* structures for the distribution.")
        n = gtotal

    result = {}
    rtotal = [0]
    def vinsert(key, value, rtotal, n, result):
        if rtotal[0] < n: #Only update if we are deficient in structures.
            if value > 0:
                if key in result:
                    result[key] += value
                else:
                    result[key] = value
                rtotal[0] += value
            return True
        else:
            return False   
    
    def assign(relvals, f, rtotal, gtotal, n, result, dataset, ftype,seed_val=None):
        """Recurses through the relative weights specified to fill the
        result dictionary until it has the exact number of structures
        requested.
        """
        if n == gtotal:
            for key, value, limit in relvals:
                vinsert(key, limit, rtotal, n, result)
        else:
            from math import modf
            from random import random, seed
            from operator import itemgetter

            if seed_val is not None:
                seed(a=seed_val)
                
            ids = {
                "all": lambda key: key,
                "shape": lambda key: key[0:2],
                "conc": lambda key: (key[0], key[2]),
                "size": lambda key: (key[0],)
            }
            #Sort the relative values according to their weights so that
            #the most abundant ones pick up the extras.
            relvals = sorted(relvals, key=itemgetter(1), reverse=True)
            recount = 0
            while rtotal[0] < n and recount < 4:
                delta = n-rtotal[0]
                for key, value, limit in relvals:
                    #If the abundance guarantees an integer part, use that, but then
                    #use a random number to decide who to keep for fractional parts.
                    #Eventually this should terminate unless all of the fractional
                    #pieces are super small; but even then, we should get what we want
                    #by the time we have iterated through the whole list.
                    rf, qf = modf(value if recount == 0 else f(delta, *(ids[ftype](key)),dataset=dataset,ftotal=float(gtotal)))
                    retval = int(qf + (1 if random() < rf else 0))
                    if not vinsert(key, retval, rtotal, n, result):
                        break
                recount += 1

            if recount == 3 and rtotal[0] < n: #pragma: no cover
                from .msg import warn
                warn("Reached maximum recursion limit in random assignment.")
                
    #First, compile a dictionary at the lowest level that has the weights
    #for the desired value of 'n'.
    relvals = []
    if ftype=="all":
        for size, data in list(dataset.items()):
            for HNF, distr in list(data["distr"].items()):
                for conc, limit in list(distr.items()):
                    key = (size, HNF, conc)
                    value = f(n, size, HNF, conc, dataset, float(gtotal))
                    relvals.append((key, value, limit))
    elif ftype == "shape":
        for size, data in list(dataset.items()):
            for HNF, limit in list(data["stotals"].items()):
                key = (size, HNF, None)
                value = f(n, size, HNF, dataset, float(gtotal))
                relvals.append((key, value, limit))                
    elif ftype == "conc":
        for size, data in list(dataset.items()):
            for conc, limit in list(data["ctotals"].items()):
                key = (size, None, conc)
                value = f(n, size, conc, dataset, float(gtotal))
                relvals.append((key, value, limit))
    elif ftype == "size":
        for size, data in list(dataset.items()):
            limit = data["gtotal"]
            key = (size, None, None)
            value = f(n, size, dataset, float(gtotal))
            relvals.append((key, value, limit))
    
    #This makes the selection according to the relative values and keeps
    #choosing, weighted by relative abundance, until we have the right
    #number of structures.
    assign(relvals, f, rtotal, gtotal, n, result, dataset, ftype, seed_val=seed)
    #Make sure we are returning exactly how many they asked for.
    if rtotal[0] > n: #pragma: no cover
        from .msg import warn
        warn("More structures were returned than asked for, should not be possible.")

    return result

def _distribution(infile, ftype, dataset, gtotal, cast=float, cellsizes=None,res_type = None,
                  res_values=None):
    """Returns the Polya-weighted distribution function for the specified cell sizes.
    If dataset is None, a dataset and total are loaded using the specified cell sizes.
    This dataset and the total number of structures it contains are returned as
    arguments 2 and 3 of the tuple. Depending on the value of 'ftype', it returns 
    a tuple (f, dataset, gtotal) where 'f' is a function function with arguments:

    'all':   f(n ,size, HNF, conc)
    'shape': f(n, size, HNF)
    'conc':  f(n, size, conc)
    'size':  f(n, size)

    where 'size' is the integer cell size; 'HNF' is the array of 6-elements defining
    the HNF matrix that is the cell shape; 'conc' is a list of concentrations such as
    [1,5] for a 6-atom cell; 'n' is the *grand* total number of structures desired from
    calling the function over all possible values of the various parameters passed to
    the distribution function.

    The value returned by the function is an integer specifying how many of that
    structure type should be selected according to the total distribution across
    all structures included in cellsizes.

    Args:
        inlfile (str): The polya.out style input file.
        dataset (dict): the dictionary returned by method: _distribution_summary().
        gtotal (int): the total number of unique structures corresponding to the
            dictionary 'dataset'.
        ftype (str): one of ['shape', 'conc', 'size', 'all']; determines the number of 
            arguments that the distr. function will accept and average over, thus controlling
            the granularity of the predictions.
        cellsizes (list): a list of the cell sizes to include in the data set summary.
        res_type (str): 'shape' if shapes are being restricted, or 'conc' if 
            concentrations are  being restricted.
        res_values (list): The allowed values of the restricted parameter.

    Returns:
        f (function): The Polya-weighted distribution function.
        dataset (dict): The dictionary returned by method: _distribution_summary().
        gtotal (int): The total number of unique structures.
    """
    if dataset is None:
        if cellsizes is not None:
            if res_type is not None:
                if res_type == "shape":
                    (dataset, gtotal) = _distribution_summary(infile, cellsizes, HNFs=res_values)
                elif res_type == "conc":
                    (dataset, gtotal) = _distribution_summary(infile, cellsizes,
                                                              wanted_concs=res_values)
                else:
                    raise ValueError("Cannot filter the distribution using {}. Please use 'shape' or 'conc'.".format(res_type))

            else:
                (dataset, gtotal) = _distribution_summary(infile, cellsizes)
        else: #pragma: no cover
            from .msg import err
            err("No dataset or cell sizes specified.")
            return None
        
    #We just need to sum up the total number of unique structures across all the
    #given cell sizes and then create a function that weights a total number of
    #desired structures by concentration and cell shape.
    ftotal = float(gtotal)
    if ftype == "all":
        def f(n,size,HNF,conc,dataset,ftotal):
            """Summing function"""
            return cast(dataset[size]["distr"][tuple(HNF)][conc]/ftotal*n)
        # f = lambda n, size, HNF, conc: cast(dataset[size]["distr"][tuple(HNF)][conc]/ftotal*n)
    elif ftype == "shape":
        def f(n,size,HNF,dataset,ftotal):
            """Summing function"""
            return cast(dataset[size]["stotals"][tuple(HNF)]/ftotal*n)
        # f = lambda n, size, HNF: cast(dataset[size]["stotals"][tuple(HNF)]/ftotal*n)
    elif ftype == "conc":
        def f(n,size,conc,dataset,ftotal):
            """Summing function"""
            return cast(dataset[size]["ctotals"][tuple(conc)]/ftotal*n)
        # f = lambda n, size, conc: cast(dataset[size]["ctotals"][tuple(conc)]/ftotal*n)
    elif ftype == "size":
        def f(n,size,dataset,ftotal):
            """Summing function"""
            return cast(dataset[size]["gtotal"]/ftotal*n)
        # f = lambda n, size: cast(dataset[size]["gtotal"]/ftotal*n)
    else:
        raise ValueError("The parameter {} is not a valid parameter for the distribution. "
                         "Please use size, shape, conc, or all.".format(ftype))

    return (f, dataset, gtotal)
        
def _distribution_summary(infile, cellsizes, HNFs = None, wanted_concs = None):
    """Returns a dictionary that summarizes the Polya predictions for unique structures
    by cell size, shape and concentration for use by a distribution function. Returns
    a tuple (summary dict, grand total), where grand total is the total number of structures
    predicted for *all* cell sizes.

    Args:
        inlfile (str): The polya.out style input file.
        cellsizes (list): a list of the cell sizes to include in the data set summary.
        HNFs (list): a list of the allowed HNFs.
        wanted_concs (list): a list of the allowed concentrations.

    Returns:
        dataset (dict): The data retreived from the polya.out style files.
        gtotal (int): The total number of structures.u
    """
    from os import path
    from numpy import loadtxt
    gtotal = 0
    dataset = {}
    for s in cellsizes:
        source = infile+"."+str(s)
        if not path.isfile(source): #pragma: no cover
            from .msg import err
            err("Cannot find polya distribution for size {} at {}".format(s, source))
            continue

        #First load the possible concentrations from the commented first line.
        with open(source) as f:
            headings = f.readline().split()
            concs = [tuple(map(int, h.split(":"))) for h in headings[2:-1]]
        polya = loadtxt(source, int)
        distr = {}

        if HNFs is None and wanted_concs is None:
            for iHNF, HNF in enumerate(polya[0:-1,0:6]):
                distr[tuple(HNF)] = {tuple(c): v for c, v in zip(concs, polya[iHNF,6:-1])}
            stotals = {tuple(m): v for m, v in zip(polya[0:-1,0:6], polya[:-1,-1])}
            ctotals = {c: v for c, v in zip(concs, polya[-1, 6:-1])}
        elif HNFs is not None:
            ctotals = {c: v for c, v in zip(concs,[0]*len(concs))}
            stotals = {}
            for iHNF, HNF in enumerate(polya[0:-1,0:6]):
                if tuple(HNF) in HNFs:
                    distr[tuple(HNF)] = {tuple(c): v for c, v in zip(concs, polya[iHNF,6:-1])}
                    for c, v in zip(concs, polya[iHNF,6:-1]):
                        ctotals[c] += v
            for m, v in zip(polya[0:-1,0:6], polya[:-1,-1]):
                if tuple(m) in HNFs:
                    stotals[tuple(m)] = v
        elif wanted_concs is not None:
            ctotals = {c: v for c, v in zip(concs,[0]*len(concs))}
            stotals = {}
            for iHNF, HNF in enumerate(polya[0:-1,0:6]):
                cs = []
                vs = []
                count = 0

                for c, v in zip(concs, polya[iHNF,6:-1]):
                    if c in wanted_concs:
                        ctotals[c] += v
                        cs.append(c)
                        vs.append(v)
                        count += 1

                distr[tuple(HNF)] = {tuple(c): v for c, v in zip(cs, vs)}
            for m, v in zip(polya[0:-1,0:6], polya[:-1,-1]):
                if tuple(m) in distr:
                    stotals[tuple(m)] = v

        dataset[s] = {
            "distr": distr,
            "concs": concs,
            "ctotals": ctotals,
            "stotals": stotals,
            "gtotal": sum([ctotals[key] for key in ctotals])
        }
        gtotal += dataset[s]["gtotal"]

    return (dataset, gtotal)    

def _print_distribution(distr, distribution, filename=None, header=True, append=False, show=False):
    """Prints the specified distribution to screen or file.

    :arg distr: the distribution returned by method:distribute().
    :arg distribution: The type of distribution, i.e., 'shape', 'size', 'conc', or 'all'.
    :arg filename: The output file name.
    :arg header: True if the header is to be included in the file.
    :arg append: True if the file is to be appended to.
    :arg show: True if the distribution is to be printed to the screen.
    """
    if show:
        from .msg import arb, cenum
        bysize = {}
        sfmt = " {0: <5d} | {1: <15} | {2: <5} | {3: <5d} "
        for key, value in list(distr.items()):
            size, HNF, conc = key
            skey = sfmt.format(size, "" if HNF is None else ' '.join(map(str, HNF)),
                               "" if conc is None else ':'.join(map(str, conc)), value)
            cols = (cenum["cwarn"], cenum["cinfo"], cenum["cgens"], cenum["cokay"])
            if (size, value) in bysize:
                bysize[(size, value)].append((skey, cols, tuple(HNF) if HNF is not None else None))
            else:
                bysize[(size, value)] = [(skey, cols, tuple(HNF) if HNF is not None else None)]

        from operator import itemgetter
        for size, value in sorted(bysize.keys()):
            for skey, cols, HNF in sorted([[(i or "") for i in x] for x in bysize[(size, value)]], key=itemgetter(2)):
                arb(skey, cols, "|")

    if filename is not None:
        #We don't worry about ordering it, just write them to file in whatever
        #order the keys are in.
        with open(filename, 'w' if not append else 'a') as f:
            if header:
                if distribution == "all":
                    f.write("# {0: <28}  {1: <10}  {2}\n".format("HNF", "Conc.", "Number"))
                elif distribution == "shape":
                    f.write("# {0: <18}  {1}\n".format("HNF", "Number"))
                elif distribution == "conc":
                    f.write("# {0: <6}  {1: <6}  {2}\n".format("Size", "Conc.", "Number"))
                else:
                    f.write("# {0: <6}  {1}\n".format("Size", "Number"))
            for key, value in list(distr.items()):
                size, HNF, conc = key
                if conc is None:
                    conc = []
                if distribution == "all":
                    f.write("  {0: <28}  {1: <10}  {2:d}\n".format(' '.join(map(str, HNF)), ' '.join(map(str, conc)), value))
                elif distribution == "shape":
                    f.write("  {0: <18}  {1}\n".format(' '.join(map(str, HNF)), value))
                elif distribution == "conc":
                    f.write("  {0: <6}  {1: <6}  {2:d}\n".format(size, ' '.join(map(str, conc)), value))
                else:
                    f.write("  {0: <6}  {1:d}\n".format(size, value))
                    
def make_enum_in(infile,distribution,outfile,number=None,sizes=None,save=True,seed=None,
                 restrict=None):
    """Makes an enum.in file if the distrubiton type is all with the
    desired number of structures. Otherwise prints the distribution
    information to the screen for the user.

    Args:
        infile (str): The name of the 'polya.out' style input file.
        n (int): The number of structures or 'all'.
        outfile (str): The name of the output file for the distribution
        save (bool): True if the data is to be saved to file.
        seed (float): The seed for the random number generator.

        distribution (list): The parameters that the distribution is over
            ('shape', 'conc', 'size', 'all').

        sizes (list): when specified, limit the distribution to these integer cell sizes;
            otherwise, look for all cell sizes we have data for.

        restrict (list): A list containing the restriction type ('shape','conc') 
            and the file of allowed values.

    Raises:
        ValueError: if the 'polya.*' file cannot be found.
    """

    from os import listdir
    
    files = listdir(".")
    if sizes is None:
        sizes = []
        form = infile

        ready = False
        for f in files:
            if form in f and f[0]==form[0]:
                sizes.append(int(f.split(".")[-1]))
                ready = True

    else:
        for i in sizes:
            sizedir = infile+'.'+str(i)
            if sizedir not in files:
                raise ValueError("Cannot find '{0}' in current directory.".format(sizedir))
        ready = True
                
    if ready == False:
        raise ValueError("The files {0} don't exist in this directory. You either "
                         "need to run the -polya option or else navigate into the "
                         "folder that contains the output {0} folders.".format(infile))

    if restrict is not None:
        # from numpy import loadtxt
        res_values = []
        with open(restrict[1],"r") as resf:
            for values in resf:
                if "#" in values:
                    pass
                elif restrict[0] == "shape":
                    res_values.append(tuple([int(i) for i in values.split()[0:6]]))
                elif restrict[0] == "conc":
                    res_values.append(tuple([int(i) for i in values.split()[1:3]]))

        distr = _distribute(infile, sizes,distribution,n=number,seed=seed,res_type=restrict[0],res_values=res_values)        
    else:
        distr = _distribute(infile, sizes,distribution,n=number,seed=seed)
        
    if distribution.lower() == "all":
        _print_distribution(distr,distribution,filename=outfile)
    else:
        _print_distribution(distr,distribution,filename= outfile if save else None,show=True)



"""Subtroutines needed to find the HNFs."""
import numpy as np

def _get_HNF_diagonals(n):
    """Finds the diagonals of the HNF that reach the target n value.
    
    Args:
        n (int): The target determinant for the HNF.
        
    Retruns:
        diags (list of lists): The allowed values of the determinant.
    """
    
    diags = []
    for i in range(1,n+1):
        if not n%i == 0:
            continue
        else:
            q = int(n//i)
            for j in range(1,q+1):
                if not q%j == 0:
                    continue
                else:
                    diags.append([i,j,q//j])
                    
    return diags

def find_all_HNFs(n):
    """Finds the HNFs for a given volume factor.
    
    Args:
        n (int): The determinant of the HNFs.

    Retruns:
        hnfs (array-like): A list of the HNFs.
    """

    diags = _get_HNF_diagonals(n)

    hnfs = []
    for diag in diags:
        a = int(diag[0])
        c = int(diag[1])
        f = int(diag[2])
        
        for b in range(c):
            for d in range(f):
                for e in range(f):
                    hnf = [[a,0,0],[b,c,0],[d,e,f]]
                    hnfs.append(hnf)

    return hnfs

def remove_duplicate_lattices(hnfs,pLV,base_vecs,base_perms,eps_=None):
    """Removes symmetrically equivalent lattices from the list of possible lattices.

    Args:
        hnfs (array-like): A list of the possible HNFs.
        pLV (array-like): The parent lattice vectors as a column matrix.
        base_vecs (array-like): The atomic basis vectors.
        base_perms (RotPermList): Permutations of the basis vectors.
        eps_ (float optional): Floating point tollerance.

    Returns:
        (uhnfs, fix_ops, rp_list, degeneracy): uhnfs is an array-like list of unique HNFs,
           fix_ops is of type OpList, i.e., the list of operations that preserve the 
           symmetry group rp_list is of type RotPermList, the permutations of basis 
           vectors, and degeneracy is a list of the degeneracy of each HNF.
    """

    from .symmetry import get_spaceGroup
    from .grouptheory import _is_equiv_lattice, _get_sLV_fixing_operations

    if eps_:
        eps = eps_
    else: #pragma: no cover
        eps = 1E-7

    a_type = [1]*len(base_vecs)

    (rots, shifts) = get_spaceGroup(pLV,a_type,base_vecs,eps=eps)
    A = np.transpose(pLV)
    
    uhnfs = []
    if len(hnfs) == 1:
        uhnfs = hnfs
    else:
        for hnf_i in hnfs:
            duplicate = False
            for hnf_j in uhnfs:
                for r in rots:
                    test_latticei = np.matmul(r,np.matmul(A,hnf_i))
                    test_latticej = np.matmul(A,hnf_j)
                    
                    if _is_equiv_lattice(test_latticei,test_latticej,eps):
                        duplicate = True
                        break
                if duplicate == True:
                    break

            if not duplicate:
                uhnfs.append(hnf_i)

    latts = []
    fix_ops = []
    rp_list = []
    degens = []
    for hnf in uhnfs:
        B = np.matmul(pLV,hnf)
        latts.append(B)
        (fix_op, RPlisti,degeneracy) = _get_sLV_fixing_operations(hnf,pLV,rots,shifts,
                                                                  base_perms,eps)
        fix_ops.append(fix_op)
        rp_list.append(RPlisti)
        degens.append(degeneracy+1)

    return (uhnfs,fix_ops,rp_list,degens)

def get_HNFs(n,pLV,base_vecs,LatDim,eps_=None):
    """Finds the unique HNFs for the system.

    Args:
        n (array-like): The determinant of the HNFs.
        pLV (array-like): The parent lattice vectors as a column matrix.
        base_vecs (array-like): The atomic basis vectors.
        LatDim (int): 3 if bulk, 2 if surf.
        eps_ (float optional): Floating point tollerance.

    Returns:
        hnfs (array-like): A list of the unique HNFS
    """

    from .grouptheory import _get_dvector_permutations
    from .symmetry import bring_into_cell
    from copy import deepcopy

    if eps_:
        eps = eps_
    else:
        eps = 1E-7

    hnfs = find_all_HNFs(n)
    temp_basis = deepcopy(base_vecs)
    plv_inv = np.transpose(np.linalg.inv(pLV))
    for i, b_vecs in enumerate(base_vecs):
        base_vecs[i] = bring_into_cell(b_vecs,plv_inv,np.transpose(pLV),eps)
        if not np.allclose(b_vecs, temp_basis[i], rtol=0, atol=eps): #pragma: no cover
            from phenum.msg import warn
            warn("An atomic basis vector was not inside the unit cell. It has been "
                 "remapped.\n Original vector: {} \n Remapped vector: {}"
                 .format(" ".join([str(p) for p in temp_basis[i]]),
                         " ".join([str(p) for p in b_vecs])))
        
    base_perms = _get_dvector_permutations(pLV,base_vecs,LatDim,eps)

    (hnfs,fix_ops,rp_list,degens) = remove_duplicate_lattices(hnfs,pLV,base_vecs,
                                                                   base_perms,eps_=eps)
    return hnfs



#!/usr/bin/env python
from phenum import msg

def RepresentsInt(s):
    """Determines if a stri could be an int.
    Args:
        s (str): The string to be tested.
    
    Returns:
        True if the string is an integer.
    """    
    try: 
        int(s)
        return True
    except ValueError:
        return False

def _make_structures(args, return_euids = False):
    """Makes a VASP POSCAR file for the desired structures.
    Args:
        args (dict): The user input.
        return_euid (bool): a unique identifier for the structures in the
            POSCARS created.
    """
    from phenum.io_utils import read_enum_out, write_POSCAR, write_config
    from phenum.vector_utils import map_enumStr_to_real_space, cartesian2direct

    (system, structure_data) = read_enum_out(args)

    euids = []
    # for each structure write the vasp POSCAR
    for structure in structure_data:
        # space_data is a dictionary containing the spacial data for
        # the structure
        euid = "".join((str(i) for i in system["plattice"]))
        euid += "".join(str(i) for i in structure["HNF"])
        euid += structure["labeling"]
        euid = (euid,structure["directions"],str(args["rattle"]),str(args["displace"]))
        euids.append(hash(euid))
        space_data = map_enumStr_to_real_space(system,structure,args["mink"])

        space_data["aBas"] = cartesian2direct(space_data["sLV"],
                                              space_data["aBas"],system["eps"])

        if args["config"]=="f":
            write_POSCAR(system,space_data,structure,args)
        elif args["config"]=="t":
            write_config(system,space_data,structure,args,args["mapping"])
            
    if return_euids: #pragma: no cover
        return euids

def examples():
    """Print some examples on how to use this python version of makeStr."""
    script = "makeStr: Makes a vasp style POSCAR for the desired system."
    explain = ("For all the examples bellow it is assumed you have already "
               "run the enumeration code and produced an enum.out style file.")
    contents = [("Make a single POSCAR file",
            "To make a POSCAR file for a specific structure listed in the "
            "`enum.out` style file you will need to identify the structure \n number "
            "(the first number of each row in the file) for the structure you want "
            ". For example to make a POSCAR for structure number 10 \n from an `enum.out` "
            "file.","makeStr.py 10 \n"),
           ("Make multilpe POSCARS at once",
            "To make multiple POSCARS for a range of values in the `enum.out` style "
            "file simply list the starting and ending structure numbers \n of the range. "
            "To make POSCARS for every structure in the output file use the word `all`.",
            "makeStr.py 10 20 \n  makeStr.py all \n"),
            ("Find the lattice parameter for the system",
             "To have makeStr.py predict the lattice parameter for the system using "
             "Vegard's Law use the -species option followed by a space \n separated list "
             "of the elements in the system.","makeStr.py 10 -species Al Cu \n"),
            ("Include displacements in POSCAR",
             "If `arrows` (displacement directions) were included in the enumeration "
             "then it is possible to displace them off the lattice points \n when making the "
             "POSCARS using the -displace option followed by the displacement amount "
             "expressed in terms of the lattice parameter. \n In other words if `a` is "
             "the lattice parameter and the atoms were to be displaced by `a/2` then "
             "the command would be:","makeStr.py 10 -displace 0.5 \n"),
            ("Make displacements have different lengths in POSCAR",
             "If `arrows` were included in the model and the `-displace` flag is being "
             "used it is possible to 'rattle' the displacements so that \n they are not all "
             "the same length. Using the `-rattle` option applies a random distribution "
             "to the displacements with the larges change \n in the displacements specified "
             "by the user as a fraction of the displacement given. So if a displacement of "
             "0.5 was given and the \n displacements were to be randomized by 1/4 of that total "
             "displacement the the command would be:",
             "makeStr.py 10 -displace 0.5 -rattle 0.25"),
            ("Make an MTP config file for a ternary endge from a binary enumeration",
             "MTP is an excellent interatomic potential capable of quickly relaxing "
             "a cell as such we would like to be able to output the structures we've "
             "selected to an MTP compatible config file rather than a VASP POSCAR. "
             "There are also times when a binary enumeration is desired to form the edges "
             "of a ternary phase diagram, to do so the species present have to be able to "
             "be mapped to other species in use. To make a binary edge of a ternary phase "
             "diagram for an MTP config file use:",
             "makeStr.py 10 -config t -species Al Cu Ni -outfile=to-relax.cfg "
             "-species_mapping 1 2 (or 0 2)")]
    required = ("REQUIRED: A `enum.out` file.")
    output = ("RETURNS: A vasp style POSCAR labeled vasp.* where the `*` is replaced "
              "with the structure number for the `enum.out` file.")
    details = ("")
    outputfmt = ("")

    msg.example(script, explain, contents, required, output, outputfmt, details)

script_options = {
    "structures": dict(nargs="+",
                        help=("The desired structure numbers from the enum.out file. This "
                              "can be either a single value or a desired range indicated by "
                              "the starting and stopping structure numbers.")),
    "-displace": dict(default=0.0, type=float,
                        help=("The displacement amount for the arrows in units of the lattice "
                               "parameter. Default is 0.")),
    "-input": dict(default="struct_enum.out",type=str,
                        help=("Override the default 'struct_enum.out' file name.")),
    "-mink": dict(default="t", choices=["t","f"],
                        help=("Sets flag to perform minkowski reduction of the basis (T/F)."
                              " Default is True.")),
    "-species": dict(default=None, nargs="+",type=str,
                        help=("Specify the atomic species present in the system.")),
    "-species_mapping": dict(default=None, nargs="+",type=int,
                        help=("Specify the atomic species numbers for the system. This option "
                              "is only used for making MTP config files.")),
    "-outfile": dict(default="vasp.{}",type=str,
                        help=("Override the default output file names: 'vasp.{structure#}'" 
                              "for the structures.")),
    "-rattle": dict(default=0.0, type=float,
                        help=("Randomizes the positions of the atoms in the POSCAR by no "
                              "more than the fraction of the displacement provided.")),
    "-config" : dict(default="f",choices=["t","f"],
                   help=("make an MTP config file instead of a VASP POSCAR. If the "
                         "MTP config file already exists it will be appended to, not "
                         "overwritten.")),
    "-remove_zeros" : dict(default="f",choices=["t","f"],
                   help=("Remove the zeros from the concentrations string in the 'POSCAR'."))    
}
"""dict: default command-line arguments and their
    :meth:`argparse.ArgumentParser.add_argument` keyword arguments.
"""

def _parser_options():
    """Parses the options and arguments from the command line."""
    #We have two options: get some of the details from the config file,
    import argparse
    from phenum import base
    pdescr = "POSCAR contstruction."
    parser = argparse.ArgumentParser(parents=[base.bparser], description=pdescr)
    for arg, options in script_options.items():
        parser.add_argument(arg, **options)
        
    args = base.exhandler(examples, parser)
    if args is None:
        return

    return args #pragma: no cover

def run(args):
    """Generates the vasp output file for the desired structure.
    Args:
        args (dict): The user input.
    """

    if args is None:
        raise ValueError("Please enter a single structure number, two structures that "
                         "indicate the first and last structure to be used in the input "
                         "file, or all.")
    elif args["structures"] != None :
        if not RepresentsInt(args["structures"][0]) and args["structures"][0].lower() == "all":
            args["structures"] = None
        elif len(args["structures"])  == 1 and RepresentsInt(args["structures"][0]):
            args["structures"] = [int(args["structures"][0])]
        elif len(args["structures"]) == 2:
            args["structures"] = list(range(int(args["structures"][0]),
                                            int(args["structures"][1])+1))
        else:
            raise ValueError("Please enter a single structure number, two structures that "
                             "indicate the first and last structure to be used in the input "
                             "file, or all. The values {} don't match this "
                             "format.".format(args["structures"]))
    else:
        raise ValueError("Please enter a single structure number, two structures that "
                         "indicate the first and last structure to be used in the input "
                         "file, or all. The values {} don't match this "
                         "format.".format(args["structures"]))

    if args["species_mapping"] is not None:
        args["mapping"] = {}
        for i in range(len(args["species_mapping"])):
            args["mapping"][i] = args["species_mapping"][i]
    else:
        args["mapping"] = None
        
    if args["remove_zeros"] == "t":
        args["remove_zeros"] = True
    else:
        args["remove_zeros"] = False

    _make_structures(args)
        
if __name__ == '__main__':
    run(_parser_options())



"""Methods for taking an HNF and finding the site permutations and the
arrow permutations. All of these have been modified for python from
their original fortran implementations which can be found at:
https://github.com/msg-byu/enumlib/. Of these only get_rotation
get_rotation_perms_lists() has been modified in order to produce the
arrow group. The class ArrowPerm has also been added and implemented
inside the RotPermList class.
"""

import itertools
import numpy as np
import operator

class ArrowPerm(object):
    """ArrowPerm pairs a site permutation with an arrow permutation.

    Attributes:
        site_perm (list): A 1D array containing a site permutation.
        arrow_perm (list): A 1D array containing an arrow permutation.
    """
    def __init__(self,site_perm = None,arrow_perm = None):
        """Initializes the ArrowPerm.
        
        Args:
            site_perm (list): A 1D array containing a site permutation.
            arrow_perm (list): A 1D array containing an arrow permutation.
        """
        self.site_perm = site_perm
        self.arrow_perm = arrow_perm

class RotPermList(object):
    """RotPermList maintains the data structure for the rotations and
    permutations from the fortran code.

    Attributes:
        nL (int): An integer indicating the number of operations.
        v (array-like): A 2D array containing the lattice vectors.
        perm (ArrowPerm): An ArrowPerm object.
        RotIndx (list): List of the rotation indices.
    """
    def __init__(self,nL = None,v =None,perm = None,RotIndx = None,arrows=None):
        """Initializes the RotPermList.

        Args:
            nL (int): An integer indicating the number of operations.
            v (array-like): A 2D array containing the lattice vectors.
            perm (ArrowPerm): An ArrowPerm object.
            RotIndx (list):  List of the rotation indices.
        """
        self.nL = nL
        self.v = v
        self.perm = ArrowPerm(site_perm = perm, arrow_perm=arrows)
        self.RotIndx = RotIndx

class OpList(object):
    """OpList maintains the data structure of the fixing operations form
    the fortran code.

    Attributes:
        rot (array-like): A 3D array containing the rotations.
        shift (array-like): A 2D array containing the translations of the lattice.
    """
    def __init__(self,rot = None, shift = None):
        """Initializes the OpList.
        
        Args:
            rot (array-like): A 3D array containing the rotations.
            shift (array-like): An 2D array containing the translations of th lattice.
        """
        self.rot = rot
        self.shift = shift

def _make_member_list(n):
    """Takes the length of three cyclic group and constructs the member
    list so that the permutations can be determined. Each member has
    three components, corresponding to the entries for each of the
    three cyclic groups.
    
    Args:
        n (int): array containing the diagonal elements of the SNF.

    Returns:
        p (list): The member list for the symmetry group.
    """
    from operator import mul
    from functools import reduce
    
    depth = int(round(reduce(mul,n,1)))
    p = np.zeros([depth,3])
    for im in range(1,depth):  # Loop over the members of the translation group
        p[im] = list(p[im-1]) # Start with the same digits as in the previous increment        
        p[im][2] = (p[im-1][2]+1)%n[2]  # Increment the first cyclic group
        if (p[im][2]==0):             # If it rolled over then
            p[im][1] = (p[im-1][1] +1) % n[1] # increment the next cyclic group
            if (p[im][1]==0):          # If this one rolled over too
                p[im][0] = (p[im-1][0]+1)%n[0] # Then increment the third one
    return p

def _find_permutation_of_group(g,gp):
    """Constructs the permutation group from the point group.
    
    Args:
        g (list): The unpermuted symmetry group.
        gp (list): The permuted symmetry group.

    Returns:
        perm (list): Permutation of g to get gp.
    """
    n = len(g)
    perm = []
    skip = [False]*n
    for im in range(n):
        for jm in range(n):
            if skip[jm]:
                continue # This is just for efficiency
            if np.allclose(gp[jm],g[im]):
                perm.append(jm)
                skip[jm] = True
                break # don't keep looking if you already found the match
            
    return perm

def _is_equiv_lattice(lat1,lat2,eps):
    """This function determines whether two lattices are equivalent. That
    is, if one is an equal-volume derivative lattice of the other. It
    uses the idea of Santoro and Mighell (Acta. Cryst. 1972) that, for
    equivalent lattices, the transformation matrix that takes the
    vectors of str2 to str1 has determinant of 1 and all integer
    elements. [LV2]=[LV1][S]=&gt;S=LV1^-1*LV2. But I had to use abs on
    the determinant. This is not mentioned by Santoro and Mighell
    (issue of opposite handedness---which doesn't matter to me).

    Args:
        lat1 (array-like): 2D array containing the first lattice.
        lat2 (array-like): 2D array containing the second lattice.
        eps (float): The finite precision tolerance.

    Returns:
        is_equiv_lattice (bool): True if lattice are equivalent.
    """
    from numpy import linalg, allclose, matmul
    is_equiv_lattice = False
    lat1inv = linalg.inv(lat1)
    S = matmul(lat1inv,lat2).tolist()
    if (allclose(abs(linalg.det(S)),1.0,rtol=0,atol=eps) and
        allclose(S,[[int(round(S[i][j])) for j in range(3)] for i in range(3)],rtol=0,atol=eps)):
        is_equiv_lattice = True
    return is_equiv_lattice
        
def _get_sLV_fixing_operations(HNF,pLV,rot,shift,dPerm,eps):
    """This subroutine deteremines which operations of the parent cell
    don't change the supercell. These operations are the symmetry
    operations of the supercell.

    Args:
        HNF (list): A 2D integer list containing the hermite normal form matrix.
        pLV (array-like): A 2D array containing the parent lattice vectors.
        rot (array-like): A 3D array containing all the rotation matrices.
        shift (array-like): A 2D array containing the translations. 
        dPerm (list): List of basis vector permutations.
        eps (float): Finite precision tolerance.

    Returns:
        fix_op (OpList): A list of the lattice fixing symmetry operations.
        rot_perm (list): The rotation permutations.
        degeneracy (int): Degeneracy of the supercell.
    """

    n_rot = len(rot)
    degen_lattices = np.zeros([n_rot,3,3])
        
    c_degen = 0
    ic = 0 # Counter for the fixing operations
    tv = 0
    tmp_op_rot = []
    tmp_op_shift = []
    tv = []
    t_index = [] # temp variables
    for iRot in range(n_rot):  # Loop over each rotation
        this_rot = rot[iRot] # Store the rotation
        orig_lat = np.matmul(np.transpose(pLV),HNF)  # Compute the superlattice
        rot_lat = np.matmul(this_rot,orig_lat) # Compute the rotated superlattice
        if _is_equiv_lattice(rot_lat,orig_lat,eps):
            # this operation fixes the lattice and should be recorded
            ic += 1
            tmp_op_rot.append(this_rot)
            tmp_op_shift.append(shift[iRot])
            tv.append(dPerm.v[iRot])
            t_index.append(iRot)
            # Added by LN from here
        else:
            in_list = False
            for iDegen in range(c_degen):
                if _is_equiv_lattice(degen_lattices[iDegen],rot_lat,eps):
                    in_list = True
                    break
          
            if not in_list:
                degen_lattices[c_degen] = rot_lat
                c_degen += 1
        # End of LN additions for degeneracy
        # Now we know which rotations fix the lattice and how many
        # there are so store them
    degeneracy = c_degen
    # Allocate the storage for them
    fix_op = OpList(tmp_op_rot,tmp_op_shift) # Stuff the rotations into the permanent array

    rot_perm = RotPermList(v=tv,RotIndx=t_index)
    
    return(fix_op,rot_perm,degeneracy)

def _map_dvector_permutation(rd,d,eps):
    """Maps the basis vectors to a permutation.
    
    Args:
        rd (array-like): 2D array of the rotated basis vectors.
        d (array-like): 2D array of the original basis vectors.
        eps (float): Finite precision tolerance.

    Returns:
       RP (list): The permutation of the basis vectors.
    """
    n_d = len(rd) # of d-vectors
    found = [False]*n_d
    RP = []
    for iD in range(n_d):
        for jD in range(n_d):
            if found[jD]:
                continue
            if np.allclose(rd[iD],d[jD],atol=eps,rtol=0):
                RP.append(jD)
                found[jD] = True
                break

    if len(RP) != len(d): #pragma: no cover
        print("d-vector didn't permute in map_dvector_permutation "
              "This usually means that the d-set from the input structure and the d-set"
              " from the struct_enum.out have a different origin or don't live in the same"
              " unit cell. This probably isn't your fault---the code should overcome this."
              ,RP)
        exit()
    return(RP)

        
def _find_minmax_indices(invec):
    """Finds the indices corresponding to the minimum and maximum values
    in an integer vector.

    Args:
        invec (list): The input array.

    Returns:
        this_min (int): The minimum index.
        this_max (int): The maximum index.
    """

    vec = [abs(i) for i in invec]
    
    this_min = vec.index(min([i for i in vec if i > 0]))

    rvec = [i for i in reversed(vec)]
    
    this_max = 2 - rvec.index(max(rvec))

    return (this_min,this_max)

def SmithNormalForm(HNF):
    """Finds the Smith Normal Form (SNF) of an HNF as well as the left and
    right transforms.
      
    Args:
        HNF (list of list): An integer matrix for which the SNF is to be found.

    Returns:

        (M, A, B) (matrix, matrix, matrix): The M is the Smith Normal Form
        matrix of the input matrix, A is the left transform matrix,
        and B is the right transform matrix. The matrices are such
        that np.dot(np.dot(A,HNF),B) = M.

    """
    from numpy import dot
    from copy import deepcopy
    
    if np.linalg.det(HNF) < 1:
        raise ValueError("SmithNormalForm routine failed because the input matrix had a "
                         "determinant less than 1.")

    A = [[0,0,0],[0,0,0],[0,0,0]]
    Mpre = deepcopy(list(HNF))
    M = [[int(x) for x in i] for i in Mpre]
    B = [[0,0,0],[0,0,0],[0,0,0]]

    if not np.allclose(M,Mpre):
        raise ValueError("The input matrix to SmithNormalForm was not integer.")
    
    for i in range(3):
        A[i][i] = 1
        B[i][i] = 1

    j = 0
    it_cnt = 0

    stop_loop = False
    while not stop_loop:
        it_cnt += 1
        if (it_cnt >=100): #pragma: no cover
            raise RuntimeError("Bad programming in SmithNormalForm")
        
        while (3-[M[0][j],M[1][j],M[2][j]].count(0)) > 1:
            (minidx,maxidx) = _find_minmax_indices([M[0][j],M[1][j],M[2][j]])
            minm = M[minidx][j]
            mult = M[maxidx][j]//minm

            M[maxidx] = list(map(operator.sub,M[maxidx],[mult*i for i in M[minidx]]))
            A[maxidx] = list(map(operator.sub,A[maxidx],[mult*i for i in A[minidx]]))
            if dot(dot(A,HNF),B).any() != np.array(M).any(): #pragma: no cover
                print("ROW: Transformation matrices didn't work")
                exit()
        if M[j][j] == 0:
            maxidx = [abs(M[0][j]),abs(M[1][j]),abs(M[2][j])].index(max([abs(M[0][j]),abs(M[1][j]),abs(M[2][j])]))
            tmprow = list(A[j])
            A[j] = list(A[maxidx])
            A[maxidx] = tmprow

            tmprow = list(M[j])
            M[j] = list(M[maxidx])
            M[maxidx] = tmprow    
        if np.dot(np.dot(A,HNF),B).any() != np.array(M).any(): #pragma: no cover
            print("ROWSWAP: Transformation matrices didn't work")
            exit()

        while (3-M[j].count(0)) >1:
            (minidx,maxidx) = _find_minmax_indices(M[j])

            minm = M[j][minidx]
            mult = M[j][maxidx]//M[j][minidx]
            for i in range(3):
                M[i][maxidx] = M[i][maxidx]-mult * M[i][minidx]
                B[i][maxidx] = B[i][maxidx]-mult * B[i][minidx]

            if np.dot(np.dot(A,HNF),B).any() != np.array(M).any(): #pragma: no cover
                print("COLS: Transformation matrices didn't work")
                exit()

        if M[j][j] < 0:
            for i in range(3):
                M[i][j] = -M[i][j]
                B[i][j] = -B[i][j]
        elif M[j][j] == 0:
            maxidx = [abs(i) for i in M[j]].index(max([abs(i) for i in M[j]]))
            for i in range(3):
                tmp = B[i][j]
                B[i][j] = B[i][maxidx]
                B[i][maxidx] = tmp 
                tmp = M[i][j]
                M[i][j] = M[i][maxidx]
                M[i][maxidx] = tmp

        if ((3-M[j].count(0)) >1) or ((3-[M[0][j],M[1][j],M[2][j]].count(0)) >1):
            continue
        if np.dot(np.dot(A,HNF),B).any() != np.array(M).any(): #pragma: no cover
            print("COLSWAP: Transformation matrices didn't work")
            exit()

        l_div = []
        for i in range(1,3):
            for k in range(1,3):
                l_div.append((M[i][k]%M[0][0] == 0))

        if j == 0 and False in l_div:
            local = [[0,0],[0,0]] 
            for i in range(1,3):
                for k in range(1,3):
                    local[i-1][k-1] = abs(M[i][k]%M[0][0])
            nondividx = local.index(max(local))
            M[0] = list(map(operator.add,M[0],M[nondividx+1]))
            A[0] = list(map(operator.add,A[0],A[nondividx+1]))
            continue

        if j == 1 and M[2][2]%M[1][1] != 0:
            M[1] = list(map(operator.add,M[1],M[2]))
            A[1] = list(map(operator.add,A[1],A[2]))
            continue
        elif j != 1:
            j = 1
            continue
        if j == 1 and (M[2][1] != 0 or M[1][2] != 0): #pragma: no cover
            # I literally could not trigger this section of code in my testse.
            continue
        stop_loop = True

    if M[2][2] < 0:
        for i in range(3):
            M[i][2]=-M[i][2]
            B[i][2]=-B[i][2]

    if np.dot(np.dot(A,HNF),B).any() != np.array(M).any(): #pragma: no cover
        print("END: Transformation matrices didn't work.")
        exit()
    check = [M[0][1],M[0][2],M[1][0],M[1][2],M[2][0],M[2][1]]
    if any(check) != 0: #pragma: no cover
        print("Not diagonal")
        exit()
    if M[1][1] % M[0][0] != 0 or M[2][2] % M[1][1] != 0: #pragma: no cover
        print("SNF conditions not met")
        exit()
        
    return(M,A,B)

def _get_dvector_permutations(par_lat,bas_vecs,LatDim,eps):
    """This routine applies the symmetry operations of the parent lattice
    to the interior points (i.e., the d-set) to see which ones are
    equivalent. Labelings of the lattice points that are contain
    permutations only of labels on equivalant sites are physically
    equivalent and therefore redundant. We use these permutations to
    eliminate those duplicate labelings

    Args:
        par_lat (list): A 2D array containing the parent lattice vectors.
        bas_vacs (list): A 2D array containing the basis vectors for the cell.
        LatDim (int): 2 if a 2D case 3 if 3D.
        eps (float): Finite precesion tolerance.

    Returns:
        drp_list (list): The basis vector permutations.
    """

    from phenum.symmetry import get_spaceGroup, bring_into_cell
    from copy import deepcopy
    
    n_d = len(bas_vecs)
    a_type = [1]*n_d

    bv_copy = deepcopy(bas_vecs)
    (rot,shift) = get_spaceGroup(par_lat,a_type,bv_copy,eps = eps)

    if LatDim==2:
        (rot,shift) = _rm_3D_operations(par_lat,rot,shift,eps)

    n_op = len(rot)
    inv_par_lat = np.linalg.inv(np.transpose(par_lat))
    nl_temp= n_op  # Number of operations that fix the parent

    # lattice (but may permute the d-vectors)
    v_temp = []
    perm_temp = []
    for iOp in range(n_op): # Try each operation in turn and see how the
        # d-vectors are permuted for each
        # Rotate each d and add the shift
        if len(bas_vecs) == 1:
            temp_b = bas_vecs[0]
        else:
            temp_b = bas_vecs
        rd_rot = np.matmul(temp_b,rot[iOp]).tolist()

        if n_d > 1:
            rd_shift = [shift[iOp]]*n_d
            rd =  [[rd_rot[i][j] + rd_shift[i][j] for j in range(len(rd_rot[i]))] for i in range(len(rd_rot))]
            
        else:
            rd = [rd_rot[i] + shift[iOp][i] for i in range(len(rd_rot))]
            
        trd = list(rd)
        if n_d > 1:
            for iD in range(n_d):
                rd[iD] = bring_into_cell(rd[iD],inv_par_lat,np.transpose(par_lat),eps)
        else:
            rd = bring_into_cell(rd,inv_par_lat,np.transpose(par_lat),eps)
            
        # The v vector is the vector that must be added (it's a lattice
        # vector) to move a rotated d-vector back into the parent cell.

        if n_d > 1:
            v_temp.append([[rd[i][j] - trd[i][j] for j in range(len(rd[i]))] for i in range(len(rd))])
        else:
            v_temp.append([[rd[i] - trd[i] for i in range(len(rd))]])
        if n_d > 1:
            perm_temp.append(_map_dvector_permutation(rd,bas_vecs,eps))
        else:
            perm_temp.append(_map_dvector_permutation([rd],[bas_vecs],eps))

    drp_list = RotPermList(nL=nl_temp,v=v_temp,perm=perm_temp)

    # I don't think we should reduce this list to a unique one. Some
    # rotations that don't permute the d's could still permute the
    # g's. So we have to keep all the d permutations, even if they
    # look redundant here.
    return(drp_list)
    
def _get_rotation_perms_lists(A,HNF,L,SNF,Op,RPlist,dperms,eps, arrows=False):
    """For each HNF, we have a list of the operations (rotations + shifts,
    if present) that leave the superlattice fixed. Given this set of
    fixing operations, make a list of the permutations on the
    d-vectors (interior points of the multilattice) effected by the
    rotations. Then sort the HNFs into blocks that have the same
    rotation permutations.

    Args:
        A (numpy ndarray): Lattice vectors of the parent lattice as rows of a matrix.
        HNF (numpy ndarray): The HNF matrix.
        L (numpy ndarray): Left transforms matrix.
        SNF (numpy ndarray): The SNF matrix.
        Op (OpList): A list of symmetry ops (rots and shifts) for the parent multilattice.
        RPlist (list of RotPermList): A list of permutations effected by the Ops.
        dperms (list): The permutation of the basis vectors.
        eps (float): Finite precision tolerance.
        arrows (bool, optional): True if arrow group is to be found as well.

    Returns:
        RPlist (list of RotPermList): The symmetry operations represented as permutations
          of a list.
    """

    # Index of the superlattices; Number of d-vectors in d set    
    n = int(round(np.linalg.det(HNF[0])))
    n_h = len(HNF)    
    n_d = np.array(dperms.perm.site_perm).shape[-1]

    skip = []
    gp = []

    arrowg = [[1,0,0],[-1,0,0],[0,1,0],[0,-1,0],[0,0,1],[0,0,-1]]
    
    tg = []
    perm = []
    ident = []
    tperms_perm = []
    temp_rperms_perm = []
    ident = np.reshape([range(n*n_d)],(n_d,n))

    len_rp = len(RPlist)
    for i in range(len_rp):
        RPlist[i].nL=0  # initialize the number
    # Make the group member list for the first SNF
    diag = [SNF[0][0][0],SNF[0][1][1],SNF[0][2][2]]
    g = np.transpose(_make_member_list(diag))

    a_t = np.transpose(A)
    a_tinv = np.linalg.inv(a_t)

    for iH in range(n_h):
        if iH > 0 and not (SNF[iH][0][0] == SNF[iH-1][0][0] and SNF[iH][1][1] == SNF[iH-1][1][1] and SNF[iH][2][2] == SNF[iH-1][2][2]):
            # Make the transform matrices for taking the g's and rotating them
            diag = [SNF[iH][0][0],SNF[iH][1][1],SNF[iH][2][2]]
            g = np.transpose(_make_member_list(diag))
    
        t_inv = np.matmul(L[iH],a_tinv)
        T = np.linalg.inv(t_inv)
        n_op = len(Op[iH].rot)
        na_op = len(arrowg)
        
        temp_rperms_perm = []
        temp_arrow_perm = []
        for iOp in range(n_op): # For each rotation, find the permutation
            dap = np.zeros(na_op)
            dgp = []
            for i in range(n):
                tt = []
                tt = np.zeros(n_d).tolist()
                dgp.append(tt)

            for iD in range(n_d): # Loop over each row in the (d,g) table
                temp1 = np.transpose([RPlist[iH].v[iOp][iD]]*n)
                temp2 = np.matmul(np.matmul(Op[iH].rot[iOp],T),g)
                rag = np.transpose(np.matmul(arrowg,Op[iH].rot[iOp]))
                rgp = np.matmul(t_inv,(-temp1+temp2))
                temp_gp = np.round(rgp).astype(int)
                temp_ag = np.round(rag).astype(int)
                if not np.allclose(rgp,temp_gp,rtol=0,atol=eps): #pragma: no cover
                    print("Transform left big fractional parts")
                    exit()

                gp = temp_gp
                ag = temp_ag
                temp_diag = np.transpose([diag]*n)
                gp = gp%temp_diag # Mod by each entry of
                # the SNF to bring into group Now that the rotated group
                # is known, find the mapping of the elements between the
                # original group and the permuted group. This is the
                # permutation.
                
                skip = [False]*n
                for im in range(n):
                    for jm in range(n):
                        if skip[jm]:
                            continue # Skip elements whose mapping is already known
                        if np.allclose(gp[:,jm],g[:,im]): # these elements
                            # map to each other The list of operations that fix
                            # the superlattice are a subset of those that fix
                            # the parent lattice. RotIndx stores the indicies
                            # of the parent lattice operations in a list with
                            # as many entries as supercell fixing operations.

                            # dperms%perm stores a list of d-vector
                            # permutations, one permutation (an n_d list) for
                            # each operation in the parent lattice symmetries
                            op_indx_in_supercell_list = RPlist[iH].RotIndx[iOp]
                            row_indx_gtable = dperms.perm.site_perm[op_indx_in_supercell_list][iD]
                            dgp[im][row_indx_gtable] = jm+iD*n
                            skip[jm] = True
                            break
                        
                # do the same thing for the arrows
                skip = [False]*na_op
                for im in range(na_op):
                    for jm in range(na_op):
                        if skip[jm]:
                            continue
                        if (ag[:,jm] == arrowg[im]).all():
                            dap[im] = jm
                            skip[jm] = True
                            break

                if np.count_nonzero(dgp == 0) > 1 or np.count_nonzero(dap == 0) > 1: #pragma: no cover
                    print("(d,g)-->(d',g') mapping failed in get_rotation_perm_lists")
                    exit()

            # Now we have the (d',g') table for this rotation. Now
            # record the permutation
            # permutation in the "long form"
            temp_rperms_perm.append(np.transpose(dgp).reshape(n_d*n).tolist()) # store
            temp_arrow_perm.append(dap)

        # nomenclature:
        # N+t = rotation (N) + fractional translation (t)  (me bethinks....)
        # r = lattice translation

        # Now that we have the permutations that are effected by N+t
        # type of rotations (for each N in the point group), we need to
        # compose them with the permutations effected by lattice
        # translations, (parent) lattice translations contained inside the
        # supercell (N+t+r, in Rod's nomenclature). Only when we have these
        # two lists, and compose them, do we have a guarantee that the
        # resulting list is actually a group. For efficiency, we reduce the
        # N+t list (remove duplicates). Rod claims that the compositions
        # (with the translations) will not have duplicates.
        if len(temp_rperms_perm) > 1 and arrows == False:
            temp_rperms_perm.sort()
            temp_rperms_perm = list(temp_rperms_perm for temp_rperms_perm, _ in itertools.groupby(temp_rperms_perm))
        elif len(temp_rperms_perm) > 1 and arrows == True:
            nr = len(temp_rperms_perm[0])
            na = len(temp_arrow_perm[0])
            len_temp_rp = len(temp_rperms_perm)
            perms = []
            for i in range(len_temp_rp):
                perms.append(list(temp_rperms_perm[i])+list(temp_arrow_perm[i]))

            perms = list(perms for perms, _ in itertools.groupby(perms))
            perms.sort()
            temp_rperms_perm = []
            temp_arrow_perm = []
            len_perms = len(perms)
            for i in range(len_perms):
                temp_rperms_perm =[p[:nr] for p in perms]
                temp_arrow_perm = [p[-na:] for p in perms]
        # The rotations permutations list is now in "alphabetical"
        # order and contains no duplicates
    
        # To get the permutations effected by the r's, we don't need
        # the r's. We can merely take the member list, the group, and
        # add each element of the group to the group itself and see
        # what permutation happens. (This part is somewhat redundant
        # with make_translation_group in labeling_related module.)
        tperms_perm = []
        for ig in range(n): # The number of r's inside the superlattice (the
            # translation perms) is the same as the index n
            temp_g = np.transpose([g[:,ig]]*n)
            tg = g+temp_g#[[g[i][j]+temp_g[i][j] for j in range(len(g[i]))] for i in range(len(g))] # Add the element to the group
            temp_diag = np.transpose([diag]*n)
            tg = tg%temp_diag#[[tg[i][j]%temp_diag[i][j] for j in range(len(tg[i]))] for i in range(len(tg))] # mod by the SNF entries to
            # bring it back to the "primitive" representation
            perm = _find_permutation_of_group(np.transpose(g),np.transpose(tg))
            temp_ident = []

            len_ident = len(ident)
            for il in range(len_ident):
                temp_ident.append([ident[il][i] for i in perm])
            tperms_perm.append(np.reshape(temp_ident,(n*n_d)).tolist())
            
        rp_list_nl = len(temp_rperms_perm)*n
        rp_list_perm_sites = [0]*rp_list_nl
        rp_list_perm_arrows = [0]*rp_list_nl
        len_t_perm = len(temp_rperms_perm)
        for it in range(n): # Loop over unique rotation
            for iOp in range(len_t_perm): # Loop over translation perms (r type)
                # perms (N+t type) Form the permutation effected by
                # composing the iOp-th one with the it-th one
                rp_list_temp = [tperms_perm[it][i] for i in temp_rperms_perm[iOp]]
                rp_list_perm_sites[iOp*n + it] = rp_list_temp
                rp_list_perm_arrows[iOp*n+it] = temp_arrow_perm[iOp]
                # ^--- Having gotten both the rotations and the
                # translation in the sections above (sort_permutations_list
                # etc...), the "operators" of the rotation (one of them is
                # R_i) and the translations (one of them is T_j) are now
                # "scrambled", i.e., (T_j) o (R_i).  Since the *first*
                # Rotation R_1 = Id, the *first* entries in the
                # RPlist(iH)%perm are equivalent to pure translations only.
                # The following entries are a combination of both.

        RPlist[iH].perm.site_perm = rp_list_perm_sites
        if arrows: 
            RPlist[iH].perm.arrow_perm = rp_list_perm_arrows
        else:
            RPlist[iH].perm.arrow_perm = [range(na_op)]*len(rp_list_perm_arrows)
        RPlist[iH].n = rp_list_nl
        
    return (RPlist)
    
def get_sym_group(par_lat,bas_vecs,HNF,LatDim,arrows=True):
    """Generates the symmetry group for a given lattice and HNF

    Args:

        par_lat (numpy ndarray): a 2D array that contains the parrent
          lattice vectors

        bas_vecs (list): A list of the atomic basis vectors for the lattice.
        HNF (numpy ndarray): The HNF matrix that defines the supercell.
        LatDim (int): 2 if a surface case case 3 if bulk.
        arrows (bool, optional): True if the arrow group is to be found as well.

    Returns:
        sym_group (RotPermList): The symmetry operations represented as permutations
          of a list.
    """

    from numpy import linalg, allclose
    from phenum.symmetry import bring_into_cell, get_spaceGroup
    from copy import deepcopy

    eps = 1E-10
    # map any atoms in the basis that aren't within the cell to be in
    # the cell
    par_lat_inv = np.transpose(linalg.inv(par_lat))
    temp_basis = deepcopy(bas_vecs)
    for i, b_vecs in enumerate(bas_vecs):
        # par_lat_inv = linalg.inv(np.transpose(par_lat))
        bas_vecs[i] = bring_into_cell(b_vecs,par_lat_inv,np.transpose(par_lat),eps)
        if not allclose(b_vecs, temp_basis[i], rtol=0, atol=eps): # pragma: no cover
            from phenum.msg import warn
            warn("An atomic basis vector was not inside the unit cell. It has been "
                 "remapped.\n Original vector: {} \n Remapped vector: {}"
                 .format(" ".join([str(p) for p in temp_basis[i]]),
                         " ".join([str(p) for p in b_vecs])))

    par_rp_list = _get_dvector_permutations(par_lat,bas_vecs,LatDim,eps)

    a_type = [1]*len(bas_vecs)
    
    (sgrots,sgshifts) = get_spaceGroup(par_lat,a_type,bas_vecs,eps)
    (fixing_ops,RPList,degeneracy) = _get_sLV_fixing_operations(HNF,par_lat,sgrots,sgshifts,
                                                                par_rp_list,eps)

    (SNF,L,R) = SmithNormalForm(deepcopy(HNF))
    sym_group = _get_rotation_perms_lists(par_lat,[HNF],[L],[SNF],[fixing_ops],[RPList],
                                          par_rp_list,eps,arrows=arrows)

    return(sym_group[0])

#a_group take the set of translations of the lattice and the set of
#rotations of the lattice paired with their effect on the arrows of
#the lattice. It makes a direct product of the transformations and the
#rotations and pairs them with the combination of th arraws. If the
#resultant operation is not in the group it then adds it to the group.
def a_group(trans,rots):
    """This subroutine combines that translations of the lattice with the
    rotatians of the lattice.

    Args:
        trans (list): A 2D array where each row is an translation
          of the lattice.

        rots (list): A 3D array. Each row's first entry is the
          site permutations and the second entry is the arrow
          permutations.

    Returns:
        groupi (list): The full symmetry group.
    """
    groupi = []
    for i in trans:
        for j in rots:
            c = []            
            c.append([j[0][i[l]] for l in range(0,len(i))])
            c.append(j[1])
            if c not in groupi:
                groupi.append(c)

    return(groupi)

def a_group_gen(trans,rots):
    """This subroutine combines that translations of the lattice with the
    rotatians of the lattice for the generators of the system.
    
    Args:
        trans (list): A 2D array where each row is an translation
          of the lattice.

        rots (list): A 3D array. Each row's first entry is the
          site permutations and the second entry is the arrow
          permutations.

    Returns:
        groupi (list): The full symmetry group.
    """
    groupi = []
    for i in trans:
        for j in rots:
            c = []            
            c.append([j[0][i[l]] for l in range(0,len(i))])
            c.append(j[1])
            if c not in groupi:
                groupi.append(c)

    groupi = _group(groupi)
    
    return(groupi)

def _group(gen): # pragma: no cover
    """This subroutine takes the generators of the group then uses them
    to form the entire group.

    Args:
        gen (list): A 2D array of the generators of the group
          where each row is a seperate generator.

    Returns:
        groupi (list): The full symmetry group.
    """
    # q is a counter to track how many group elements we've found
    q = 1
    # the final output group
    groupi = []
    # loop over the generators
    for i in gen:
        # a second set of generators used to ensure that a group
        # action doesn't act an its self
        gen2 = []      
        # if k is not the ith element of the generators add it to the
        # set gen2
        for k in gen:
            if k != i:
                gen2.append(k)
        # loop over all except the ith generator
        for j in gen2:
            # handle the site group and arrow group independently
            trans = [j[0][i[0][l]] for l in range(0,len(i[0]))]
            rots = [j[1][i[1][l]] for l in range(0,len(i[1]))]
            # c is the combination of the trans and rots, if it's not
            # in groupi then it should be added.
            c = [trans,rots]
            if c not in groupi:
                groupi.append(c)

            # for each new group element, c, we should also see if
            # the ith group element acting on it will make another
            # new group element.
            trans = [c[0][i[0][l]] for l in range(0,len(i[0]))]
            rots = [c[1][i[1][l]] for l in range(0,len(i[1]))]
            # d is the group action after the ith element has acted on
            # c, if it's not in groupi then add it.
            d = [trans, rots]
            if d not in groupi:
                groupi.append(d)
            # until d and c are the same again we want to apply the
            # ith group element to d so that we've found all possible
            # group elements for the combination
            while d != c:
                trans = [d[0][i[0][l]] for l in range(0,len(i[0]))]
                rots = [d[1][i[1][l]] for l in range(0,len(i[1]))]
                d = [trans,rots]
                # increment q because we found a new group element
                q += 1
                # every time d is not in group i add it
                if d not in groupi:
                    groupi.append(d)
    # new is a tracker that is 0 while we are still finding new group
    # elements and 1 when no new ones are found
    new = 0

    #Here each generator is applied to the group over and over until no new group
    #elements are found
    while new == 0:
        # a second set to store any additional group elements
        group2 = []
        for i in gen:
            for h in groupi:
                # apply the ith generator to the hth element of groupi
                # from above to search for new group elements
                trans = [h[0][i[0][l]] for l in range(0,len(i[0]))]
                rots = [h[1][i[1][l]] for l in range(0,len(i[1]))]
                
                d = [trans,rots]
                # if do isn't in groupi or in group2 then add it to
                # group2
                if d not in groupi and d not in group2:
                    group2.append(d)
        # add anything that is in group2 to groupi
        if len(group2) > 0:
            groupi.extend(group2)
        else:
        # if no new group elements were found then we'v found the
        # whole group and new is set to 1
            new = 1
        
    return(groupi)

def get_full_HNF(HNF):
    """Turns the reduced HNF to a full HNF matrix.

    Args:
        HNF (list): The 1D array that contains all the lower
          diagonal entries of the HNF matrix.

    Returns:
        full_hnf (list of list): A 2D list containing the HNF matrix.
    """

    if type(HNF).__module__ == np.__name__:
        temp_hnf = HNF.tolist()
    else:
        temp_hnf = HNF
    full_hnf = [[temp_hnf[0],0,0],[temp_hnf[1],temp_hnf[2],0],temp_hnf[3:]]

    return full_hnf

def _rm_3D_operations(aVecs,sgrots,sgshifts,eps):
    """This subroutine removes operations that are 3 dimmensional.

    Args:
        aVecs (list): The 2D array of the primitive real space lattice vectors.
        sgrots (list): The 3D array containing the space group rotations.
        sgshifts (list): The 2D array containing the space group shifts.
        eps (float): Finite precisions tolerance.

    Returns:
        t_sg_rots (list): The 3D array of the surface rotations.
        t_sg_shifts (list): The 2D array of the surface translations.
    
    Raises:
        ValueError: if the input vectors aren't primitive.
    """
  
    if not np.allclose(
            aVecs[0][1:3],0.0,rtol=0, atol=eps) or not np.allclose(aVecs[2][0:2]
                                                                   ,0.0,rtol=0,atol=eps):
        raise ValueError("Error in rm_3d_operations: only allowed for primitive vectors x00,0xx,0xx")
    
    n_rot = len(sgrots)
    irot = 0
    t_sg_rots = []
    t_sg_shifts = []
    for i in range(n_rot):
        if (np.allclose(sgrots[i][0][1:3],0.0,rtol=0,atol=eps) and
            np.allclose([sgrots[i][1][0],sgrots[i][2][0]],0.0,rtol=0,atol=eps)
            and np.allclose(abs(sgrots[i][0][0]),1.0,rtol=0,atol=eps)):
            # this operation is "2D"         
            irot += 1
            t_sg_rots.append(sgrots[i])
            t_sg_shifts.append(sgshifts[i])

    return(t_sg_rots,t_sg_shifts)



"""Methods used for generating the symmetry group. All of these
methods were adapted to python from their original fortran
implementations which can be found at:
https://github.com/msg-byu/symlib
"""

from copy import deepcopy
import numpy
import math

def get_concs_for_size(size,nspecies,res_concs,nB,concs):
    """Gets the concentration ranges for the atoms within the cells of
    certain sizes given the constraints provided such as
    concentration restrictions and the presence of arrows. Code
    rewritten from the get_conetration_list subroutine of:
    https://github.com/msg-byu/enumlib/blob/master/src/derivative_structure_generator.f90  

    Args:
        size (int): The cell size in integer form
        nspecies (int): the number of atomic species in the system
        nB (int): the number of basis vectors being used

        res_concs (bool): a logical that indicates of the concentrations
          are being restricted

        concs (list of int): A 2D integer array that contains the concentration
          ranges for each atom in the system

    Returns:
        c_list (array-like): The allowed concentration ranges.
    """
    eps = 1E-10
    
    from itertools import product
    if res_concs == True:
        denom = concs[0][2]
        vol_table = []
        for atom in concs:
            minc = min(atom[0:2])
            maxc = max(atom[0:2])
            vol_table.append([int(math.floor(float(minc)/denom*size*nB)),int(math.ceil(float(maxc)/denom*size*nB)),size*nB])

        n = vol_table[0][2]
        digit = [vol_table[i][1]-vol_table[i][0] for i in range(len(vol_table))]
        dig_cnt = [0*i for i in range(len(vol_table))]
        k = len(vol_table)

        label = []
        minv = []
        maxv = []
        for i in range(k):
            label.append(list(range(vol_table[i][0],vol_table[i][1]+1)))
            minv.append(float(min([concs[i][0],concs[i][1]]))/concs[i][2])
            maxv.append(float(max([concs[i][0],concs[i][1]]))/concs[i][2])

        a = [label[i][0] for i in range(len(label))]
        
        c_list = []
        done = False
        while done == False:
            if sum(a) == n:
                conc = []
                len_a = len(a)
                for i in range(len_a):
                    conc.append(a[i]/float(n))
                if not ((any(conc[i] < (minv[i]-eps) for i in range(len(minv)))) or (any(conc[i] > (maxv[i]+eps) for i in range(len(maxv))))):
                    c_list.append(deepcopy(a))
            j = k-1
            done2 = False
            while done2 == False:
                if dig_cnt[j] != digit[j]:
                    done2 = True
                    break
                a[j] = label[j][0]
                dig_cnt[j] = 0
                j -= 1
                if j < 0:
                    done2 = True
                    break
            if j < 0:
                done = True
                break
            dig_cnt[j] += 1
            a[j] = label[j][dig_cnt[j]]
            
    else:
        c_list = []
        crange = list(range(0,(size*nB)+1))
        aranges = []
        for i in range(nspecies):
            aranges.append(crange)
        for p in product(*aranges):
            if sum(p) == size*nB:
                c_list.append(list(p))
    return(c_list)

def _does_mapping_exist(v,this_type,atom_pos,atomType,eps):
    """Checks to see if a mapping exists between the vector v and the
    position of any of the atoms of type "this_type".  If a mapping
    exists, then the logical "mapped" is returned .true., otherwise
    .false.

    Args:
        v (list of float): Array of the position to check mapping for
        this_type (int): Integer that indicates which type of atom that is
          being checked.

        atom_pos (array-like): 2D array of the positions of the basis
          atoms.

        atomType (list of int): Array of integers of the types of atoms in the
          basis.

        eps (float): Epsilon for checking equivalence.

    Returns:
        mapped (bool): True if mapping exists.
    """
    mapped = False
    for i, a_type in enumerate(atomType):
        if a_type == this_type:
            # if the coordinates are the same, 
            # their difference will be zero for every component
            this_position = atom_pos[i]
            if(numpy.allclose(v, this_position, rtol=0,atol=eps)):
                mapped = True
                break
    return mapped


def _get_transformations(par_lat):

    """This routine generates the matrices for converting vectors from
    lattice coordinates to cartesion coordinates and vice versa.

    Args:
        par_lat (array-like): A 2D array that contains the parent 
          lattice vectors

    Returns:
        prim_to_cart (numpy ndarray): The matrix that transforms from lattice to 
          cartesian coordinates.
        
        cart_to_prim (numpy ndarray): The matrix that tranforms form cartesian to
          lattice coordinates.
    """

    prim_to_cart = numpy.transpose(deepcopy(par_lat))
    cart_to_prim = numpy.linalg.inv(numpy.array(prim_to_cart))

    return(prim_to_cart,cart_to_prim)

def bring_into_cell(vec,cart_to_latt,latt_to_cart,eps):
    """This subroutine translates a point back into the unit cell in
    lattice coordinates, the coefficients of the point must all be
    less than one and at least zero.
    
    Args:
        vec (list of int): The atom's position vector

        cart_to_latt (numpy ndarray): The matrix that transforms from cartesian 
          to lattice coordinates

        latt_to_cart (numpy ndarray): The matrix that transforms form lattice to 
          cartesian coordinates

        eps (float): Finite precision tolerance.

    Returns:
        vec (list of float): The vector brought back into the cell.
    """

    from numpy import matmul
    # Put the representation of the point into lattice coordinates
    vec = matmul(cart_to_latt,vec).tolist()
    # counter to catch compiler bug
    c = 0
    maxc = max(math.ceil(abs(max(vec))),math.ceil(abs(min(vec))))*2

    # If a component >= 1, translate by subtracting a lattice vector
    # If a component < 0, translate by adding a lattice vector
    while any(i > 1.0-eps for i in vec) or any(i < 0.0-eps for i in vec):
        c = c +1
        for i, v in enumerate(vec):
            if v >= 1.0-eps:
                vec[i] -= 1
            elif v < 0.0-eps:
                vec[i] += 1
        if (c>maxc): #pragma: no cover
            print("ERROR: loop does not end in bring_into_cell. Probably compiler bug.")
            exit()

    # Put the point back into cartesion coordinate representation
    vec = matmul(latt_to_cart,vec).tolist()
    return vec

def get_lattice_pointGroup(a_vecs, eps=1E-10):
    """This routine returns only the point group of the lattite rather
      than the space group of the given crystal structure.

    Args:
        a_vecs (array-like): The 2D array that contains the parent lattice 
            vectors as row vectors.

        eps (float, optional): Finite precision tolerance

    Returns:
        lattpg_op (array-like): The point group for the lattice in 
        cartesian coordinates.
    """

    inverse_avecs = numpy.linalg.inv(numpy.array(a_vecs))
    # Store the norms of the three lattice vectors
    norm_avecs = []
    for i in range(3):
        norm_avecs.append(numpy.linalg.norm(a_vecs[i]).tolist())

    # Decide how many lattice points to look in each direction to get all the 
    # points in a sphere that contains all of the longest _primitive_ vectors
    cell_volume = abs(numpy.dot(a_vecs[0],numpy.cross(a_vecs[1],a_vecs[2])))
    max_norm = max([numpy.linalg.norm(i) for i in a_vecs])
    n1 = math.ceil(max_norm*numpy.linalg.norm(numpy.cross(a_vecs[1],a_vecs[2])/cell_volume+eps))
    n2 = math.ceil(max_norm*numpy.linalg.norm(numpy.cross(a_vecs[2],a_vecs[0])/cell_volume+eps))
    n3 = math.ceil(max_norm*numpy.linalg.norm(numpy.cross(a_vecs[0],a_vecs[1])/cell_volume+eps))

    r_vecs = []
    r_lengths = []

    a_vecs = numpy.array(a_vecs)
    # Store the R vectors that lie within the sphere
    num_rs = 0
    for i in range(-int(round(n1)), int(round(n1))+1):
        for j in range(-int(round(n2)), int(round(n2))+1):
            for k in range(-int(round(n3)), int(round(n3))+1):
                this_vector = i*a_vecs[0] + j*a_vecs[1] + k*a_vecs[2]
                length = numpy.linalg.norm(this_vector)
                if (length > max_norm + eps):
                    continue # This vector is outside sphere
                num_rs += 1
                r_vecs.append(this_vector.tolist())
                r_lengths.append(length)
    # Try all R vector triplets in the sphere and see which ones are valid 
    # rotations of the original basis vectors.
    # 
    # The length of all vectors must be preserved under a unitary
    # transformation so skip any trial vectors that aren't the same
    # length as the original. We also skip any set of vectors that
    # have the right lengths but do not form a parallelpiped that has
    # the same volume as the original set. Also, note that the we skip
    # sets of vectors that contain the same vector more than once
    # (i.e. the indices i, j, k must be unique).
    num_ops = 0
    lattpg_op = []
    from itertools import permutations
    for i,j,k in permutations(range(num_rs),3):
        if (abs(r_lengths[i] - norm_avecs[0]) > eps) or (abs(r_lengths[j] - norm_avecs[1]) > eps) or (abs(r_lengths[k] - norm_avecs[2]) > eps) or (abs(cell_volume - abs(numpy.linalg.det([r_vecs[i],r_vecs[j],r_vecs[k]]))) > eps):
            continue
        # Form the new set of "rotated" basis vectors
        new_vectors = [r_vecs[i],r_vecs[j],r_vecs[k]]
        # If the transformation matrix that takes the original set to the new set is
        # an orthogonal matrix then this rotation is a point symmetry of the lattice.
        rotation_matrix = numpy.matmul(inverse_avecs,new_vectors)
        # Check orthogonality of rotation matrix by [R][R]^T = [1]
        test_matrix = numpy.matmul(rotation_matrix,numpy.transpose(rotation_matrix))
        if (numpy.allclose(test_matrix, [[1,0,0],[0,1,0],[0,0,1]], rtol=0,atol=eps)): # Found valid rotation
            num_ops +=  1 # Count number of rotations
            lattpg_op.append(rotation_matrix.tolist())

    return(lattpg_op)
    
def get_spaceGroup(par_lat,atomType,bas_vecs,eps=1E-10,lattcoords = False):
    """This routine takes a crystal structure (basis vectors and basis
    atoms and positions) and returns the point operators and
    fractional translations of the space group. The routine assumes
    that the given crystal structure is already primitive.

    Args:
        par_lat (array-like): A 2D array that contains the parent lattice vectors.
        atomType (list of int): Integer array representing the type of each basis atom.
        bas_vecs (array-like): A 2D array that contains the basis vectors for the cell.
        eps (float, optional): Finite precisions tolerance.
        lattcoords (bool, optional): True if vectors are in lattice coordinates 
            rather than cartesian.

    Returns:
        (sg_ops, sg_fracts) (array-like, array-like): The rotation and mirror operations of 
            the space group, and the translation operations of the space group.
    """
    # Get number of atoms in the basis    
    n_atoms = len(atomType)

    # save original atomic input positions
    atom_pos = deepcopy(bas_vecs)

    # A vector can be represented as either a set of cartesian coordi-
    # nates or as a linear combination of primitive lattice vectors
    # Get transformation matrices to take us back and forth
    (latt_to_cart,cart_to_latt) = _get_transformations(par_lat)
    
    # If we're in lattice coordinates Convert the position of the
    # basis atoms from lattice coordinates.
    if lattcoords:
        for i in range(n_atoms):
            atom_pos[i] = numpy.matmul(latt_to_cart,atom_pos[i]).tolist()

    # bring all the basis atoms into the unit cell
    for i, a_pos in enumerate(atom_pos):
        atom_pos[i] = bring_into_cell(a_pos,cart_to_latt,latt_to_cart,eps)

    # Now find the point group
    lattpg_op = get_lattice_pointGroup(par_lat,eps=eps)

    # **** Find the elements of the space group ****
    # Count the elements
    sgop_count = 0
    sg_ops = []
    sg_fracts = []
    # Apply each of the point operators in the point group to the crystal
    for op in lattpg_op:
        # rotate atom 1 and store its position in the vector v
        v = numpy.matmul(op,atom_pos[0])
        # Loop over all possible fractional translations
        for jAtom in range(n_atoms):
            if (atomType[jAtom] != atomType[0]):
                continue #pragma: no cover
            fract = [atom_pos[jAtom][i] - v[i] for i in range(3)]
            fract = bring_into_cell(fract, cart_to_latt, latt_to_cart, eps)
            # Is each atom of every type mapped by this rotation + translation?
            for kAtom in range(n_atoms):
                this_type = atomType[kAtom]
                # Rotate and translate each atom        
                v2 = numpy.matmul(op,atom_pos[kAtom])
                v2 = [v2[i] + fract[i] for i in range(3)]
                v2 = bring_into_cell(v2, cart_to_latt, latt_to_cart, eps)
                
                # Try to map this rotated atom onto another the same type
                mapped = _does_mapping_exist(v2, this_type, atom_pos, atomType, eps)
                if not mapped:
                    break # no mapping for this atom

            # if all atoms mapped successfully then count this
            # element (rotation + translation)
            if mapped:
                sgop_count += 1 # Count the number of elements
                sg_fracts.append(fract) # Save the translational part
                sg_ops.append(op) # Store the rotational part
                # loop over fractional translations and try next op
                # By removing the preceding exit, we include fractional translations
                # for non-primitive lattices. (GLWH 10/26/2009)
                
    return(sg_ops,sg_fracts)
